# -*- coding: utf-8 -*-

"""

  redis model adapter
  ~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import json
import datetime
import collections

# adapter API
from . import abstract
from .abstract import (IndexedModelAdapter,
                       DirectedGraphAdapter)

# canteen util
from canteen.util import struct
from canteen.util import decorators


## Globals
_support = struct.WritableObjectProxy()
_mock_redis = None  # holds global singleton for mock testing
_server_profiles = {}  # holds globally-configured server profiles
_default_profile = None  # holds the default redis instance mapping
_client_connections = {}  # holds instantiated redis connection clients
_profiles_by_model = {}  # holds specific model => redis instance mappings
_SERIES_BASETYPES = (  # basetypes that should be stored as a sorted set
  datetime.datetime, datetime.date, float)


##### ==== runtime ==== #####

# resolve redis
try:
  ## force absolute import to avoid infinite recursion
  redis = _redis_client = _support.redis = (
    __import__('redis', locals(), globals(), [], 0))
except ImportError:  # pragma: no cover
  _support.redis, _redis_client, redis = False, None, None

# or fakeredis, for testing only
try:
  import fakeredis; _support.fakeredis = True
except ImportError:  # pragma: no cover
  fakeredis, _support.fakeredis = None, False

# resolve gevent
try:
  import gevent; _support.gevent = gevent
except ImportError:  # pragma: no cover
  gevent, _support.gevent = None, False
else:  # pragma: no cover
  if _support.redis and (
    hasattr(redis.connection, 'socket') and hasattr(gevent, 'socket')):
    ## with Redis AND gevent, patch the connection socket / pool
    redis.connection.socket = gevent.socket


##### ==== serializers ==== #####

# resolve msgpack
try:
  import msgpack; _support.msgpack = msgpack
except ImportError:  # pragma: no cover
  msgpack, _support.msgpack = None, False


##### ==== compressors ==== #####

# resolve zlib
try:
  import zlib; _support.zlib = zlib
except ImportError:  # pragma: no cover
  zlib, _support.zlib = None, False

# resolve snappy
try:
  import snappy; _support.snappy = snappy
except ImportError:  # pragma: no cover
  snappy, _support.snappy = None, False

# resolve lz4
try:
  import lz4; _support.lz4 = lz4
except ImportError:  # pragma: no cover
  lz4, _support.lz4 = None, False


class RedisMode(object):

  """ Map of hard-coded modes of internal operation for the `RedisAdapter`. """

  hashkey_hash = 'hashkey'  # HSET <key>, <field> => <value> [...]
  hashkey_blob = 'hashblob'  # HSET <entity_group>, <key_id>, <entity>
  hashkind_blob = 'hashkind'  # HSET <kind>, <key_id>, <entity>
  toplevel_blob = 'toplevel'  # SET <key>, <entity>


class RedisAdapter(DirectedGraphAdapter):

  """ Adapt model classes to Redis. """

  __testing__ = False  # are we testing redis?

  # key encoding
  adapter = _redis_client
  connection_spec = None

  # magic string identifiers
  _id_prefix = '__id__'
  _meta_prefix = '__meta__'
  _kind_prefix = '__kind__'
  _magic_separator = '::'
  _path_separator = '.'
  _chunk_separator = ':'


  class EngineConfig(object):

    """ Configuration for the `RedisAdapter` engine. """

    encoding = True  # encoding for keys and special values
    serializer = json  # json or msgpack
    compression = False  # compression for serialized data values
    mode = RedisMode.toplevel_blob  # internal mode of operation


  class Operations(object):

    """ Available datastore operations. """

    ## Key Operations
    SET = 'SET'
    GET = 'GET'
    KEYS = 'KEYS'
    DUMP = 'DUMP'
    DELETE = 'DEL'
    EXISTS = 'EXISTS'
    EXPIRE = 'EXPIRE'
    EXPIRE_AT = 'EXPIREAT'
    MIGRATE = 'MIGRATE'
    MOVE = 'MOVE'
    GETBIT = 'GETBIT'
    GETSET = 'GETSET'
    GETRANGE = 'GETRANGE'
    OBJECT = 'OBJECT'
    PERSIST = 'PERSIST'
    PEXPIRE = 'PEXPIRE'
    PEXPIREAT = 'PEXPIREAT'
    PTTL = 'TTL'
    RANDOM = 'RANDOMKEY'
    RENAME = 'RENAME'
    RENAMENX = 'RENAMENX'
    RESTORE = 'RESTORE'
    SORT = 'SORT'
    TTL = 'TTL'
    TYPE = 'TYPE'
    SCAN = 'SCAN'

    ## Counter Operations
    INCREMENT = 'INCR'
    DECREMENT = 'DECR'
    INCREMENT_BY = 'INCRBY'
    DECREMENT_BY = 'DECRBY'
    INCREMENT_BY_FLOAT = 'INCRBYFLOAT'

    ## Hash Operations
    HASH_SET = 'HSET'
    HASH_GET = 'HGET'
    HASH_KEYS = 'HKEYS'
    HASH_SCAN = 'HSCAN'
    HASH_DELETE = 'HDEL'
    HASH_LENGTH = 'HLEN'
    HASH_VALUES = 'HVALS'
    HASH_EXISTS = 'HEXISTS'
    HASH_GET_ALL = 'HGETALL'
    HASH_SET_SAFE = 'HSETNX'
    HASH_MULTI_GET = 'HMGET'
    HASH_MULTI_SET = 'HMSET'
    HASH_INCREMENT = 'HINCRBY'
    HASH_INCREMENT_FLOAT = 'HINCRBYFLOAT'

    ## String Commands
    APPEND = 'APPEND'
    STRING_LENGTH = 'STRLEN'

    ## List Operations
    LIST_SET = 'LSET'
    LEFT_POP = 'LPOP'
    RIGHT_POP = 'RPOP'
    LEFT_PUSH = 'LPUSH'
    RIGHT_PUSH = 'RPUSH'
    LEFT_PUSH_X = 'LPUSHX'
    RIGHT_PUSH_X = 'RPUSHX'
    LIST_TRIM = 'LTRIM'
    LIST_INDEX = 'LINDEX'
    LIST_RANGE = 'LRANGE'
    LIST_LENGTH = 'LLEN'
    LIST_REMOVE = 'LREM'
    BLOCK_LEFT_POP = 'BLPOP'
    BLOCK_RIGHT_POP = 'BRPOP'

    ## Set Operations
    SET_ADD = 'SADD'
    SET_POP = 'SPOP'
    SET_MOVE = 'SMOVE'
    SET_DIFF = 'SDIFF'
    SET_UNION = 'SUNION'
    SET_REMOVE = 'SREM'
    SET_MEMBERS = 'SMEMBERS'
    SET_INTERSECT = 'SINTER'
    SET_IS_MEMBER = 'SISMEMBER'
    SET_DIFF_STORE = 'SDIFFSTORE'
    SET_CARDINALITY = 'SCARD'
    SET_UNION_STORE = 'SUNIONSTORE'
    SET_RANDOM_MEMBER = 'SRANDMEMBER'
    SET_INTERSECT_STORE = 'SINTERSTORE'

    ## Sorted Set Operations
    SORTED_ADD = 'ZADD'
    SORTED_RANK = 'ZRANK'
    SORTED_RANGE = 'ZRANGE'
    SORTED_SCORE = 'ZSCORE'
    SORTED_COUNT = 'ZCOUNT'
    SORTED_REMOVE = 'ZREM'
    SORTED_CARDINALITY = 'ZCARD'
    SORTED_UNION_STORE = 'ZUNIONSTORE'
    SORTED_INCREMENT_BY = 'ZINCRBY'
    SORTED_INDEX_BY_SCORE = 'ZREVRANK'
    SORTED_RANGE_BY_SCORE = 'ZRANGEBYSCORE'
    SORTED_INTERSECT_STORE = 'ZINTERSTORE'
    SORTED_MEMBERS_BY_INDEX = 'ZREVRANGE'
    SORTED_MEMBERS_BY_SCORE = 'ZREVRANGEBYSCORE'
    SORTED_REMOVE_RANGE_BY_RANK = 'ZREMRANGEBYRANK'
    SORTED_REMOVE_RANGE_BY_SCORE = 'ZREMRANGEBYSCORE'

    ## Pub/Sub Operations
    PUBLISH = 'PUBLISH'
    SUBSCRIBE = 'SUBSCRIBE'
    UNSUBSCRIBE = 'UNSUBSCRIBE'
    PATTERN_SUBSCRIBE = 'PSUBSCRIBE'
    PATTERN_UNSUBSCRIBE = 'PUNSUBSCRIBE'

    ## Transactional Operations
    EXEC = 'EXEC'
    MULTI = 'MULTI'
    WATCH = 'WATCH'
    UNWATCH = 'UNWATCH'
    DISCARD = 'DISCARD'

    ## Scripting Operations
    EVALUATE = 'EVAL'
    EVALUATE_STORED = 'EVALSHA'
    SCRIPT_LOAD = ('SCRIPT', 'LOAD')
    SCRIPT_KILL = ('SCRIPT', 'KILL')
    SCRIPT_FLUSH = ('SCRIPT', 'FLUSH')
    SCRIPT_EXISTS = ('SCRIPT', 'EXISTS')

    ## Connection Operations
    ECHO = 'ECHO'
    PING = 'PING'
    QUIT = 'QUIT'
    SELECT = 'SELECT'
    AUTHENTICATE = 'AUTH'

    ## Server Operations
    TIME = 'TIME'
    SYNC = 'SYNC'
    SAVE = 'SAVE'
    INFO = 'INFO'
    DEBUG = ('DEBUG', 'OBJECT')
    DB_SIZE = 'DBSIZE'
    SLOWLOG = 'SLOWLOG'
    MONITOR = 'MONITOR'
    SLAVE_OF = 'SLAVEOF'
    SHUTDOWN = 'SHUTDOWN'
    FLUSH_DB = 'FLUSHDB'
    FLUSH_ALL = 'FLUSHALL'
    LAST_SAVE = 'LASTSAVE'
    CONFIG_GET = ('CONFIG', 'GET')
    CONFIG_SET = ('CONFIG', 'SET')
    CLIENT_KILL = ('CLIENT', 'KILL')
    CLIENT_LIST = ('CLIENT', 'LIST')
    CLIENT_GET_NAME = ('CLIENT', 'GETNAME')
    CLIENT_SET_NAME = ('CLIENT', 'SETNAME')
    CONFIG_RESET_STAT = ('CONFIG', 'RESETSTAT')
    BACKGROUND_SAVE = 'BGSAVE'
    BACKGROUND_REWRITE = 'BGREWRITEAOF'

  @classmethod
  def __repr__(cls):  # pragma: no cover

    """ Generate a pleasant string representation of the currently-active
        ``RedisAdapter``.

        :returns: Pretty string with config info. """

    return "%s(mode=%s, serializer=%s, compression=%s)" % (
              cls.__name__,
              cls.EngineConfig.mode,
              cls.EngineConfig.serializer and (
                cls.EngineConfig.serializer.__name__),
              cls.EngineConfig.compression and (
                cls.EngineConfig.compression.__name__))

  __unicode__ = __str__ = __repr__

  @classmethod
  def is_supported(cls):

    """ Check whether this adapter is supported in the current environment.

        :returns: The imported ``Redis`` driver, or ``False`` if it could not
          be found. """

    if cls.__testing__ and fakeredis:  # pragma: no cover
      return _support.fakeredis
    return _support.redis

  @decorators.classproperty
  def serializer(cls):  # pragma: no cover

    """ Load and return the optimal serialization codec.

        :returns: Currently-configured serializer, mounted statically at
          ``cls.EngineConfig.serializer``. """

    return msgpack if _support.msgpack else json

  @decorators.classproperty
  def compressor(cls):  # pragma: no cover

    """ Load and return the optimal data compressor.

        :returns: Currently-configured compressor, mounted statically at
          ``cls.EngineConfig.compression``. """

    return cls.EngineConfig.compression or zlib

  @classmethod
  def acquire(cls, name, bases, properties):

    """  """

    return super(RedisAdapter, cls).acquire(name, bases, properties)

  def __init__(self):

    """  """

    global _server_profiles
    global _default_profile
    global _profiles_by_model

    ## Resolve default
    servers = self.config.get('servers', False)

    if not _default_profile:

      ## Resolve Redis config
      if servers:  # pragma: no cover
        for name, config in servers.items():
          if name == 'default' or (config.get('default', False) is True):
            _default_profile = name
          elif not _default_profile:  # pragma: no cover
            _default_profile = name
          if isinstance(config, basestring):
            _server_profiles[name] = redis.from_url(config)
          else:
            _server_profiles[name] = config

      if not _default_profile:
        # still no default? inject sensible defaults
        _default_profile = '__default__'
        _server_profiles['__default__'] = {
          'host': '127.0.0.1', 'port': 6379}

  @classmethod
  def channel(cls, kind):

    """ Retrieve a write channel to Redis.

        :param kind: String :py:class:`model.Model` kind to retrieve a channel
          for.

        :raises RuntimeError:

        :returns: Acquired ``Redis`` client connection, potentially specific to
          the handed-in ``kind``. """

    global _mock_redis

    if not cls.adapter or not _support.redis:  # pragma: no cover
      raise RuntimeError('No support detected in the current environment'
                         ' for Python Redis. Please `pip install redis`.')

    if not (__debug__ and cls.__testing__):  # pragma: no cover

      impl, pool, profile = (
        cls.adapter.StrictRedis, cls.adapter.ConnectionPool, None)

      # convert to string kind if we got a model class
      if not isinstance(kind, basestring) and kind is not None:
        kind = kind.kind()

      # check for existing connection
      if kind in _client_connections:
        return _client_connections[kind]  # return cached connection

      # check kind-specific profiles
      if kind in _profiles_by_model.get('index', set()):  # pragma: no cover
        profile = _profiles_by_model['map'].get(kind)
        if 'unix_socket_path' in profile:
          client = _client_connections[kind] = (
            impl(**profile))
          return client

        max_connections = profile.get('max_connections', 1000)
        if 'max_connections' in profile:
          del profile['max_connections']
        client = _client_connections[kind] = pool(
          max_connections=max_connections, **profile)
        return impl(connection_pool=client)

      # # @TODO(sgammon): patch client with connection/workerpool (if gevent)

      # check for cached default connection
      if '__default__' in _client_connections:
        return _client_connections['__default__']

      _config = cls.config
      if 'servers' in _config:
        profile = _config['servers'].get('default', None)
        profile = _config['servers'].get(profile, None)
        if profile: return impl(**profile)

      # otherwise, build new default
      default_profile = profile = _server_profiles[_default_profile]
      if isinstance(default_profile, basestring):
        # if it's a string, it's a pointer to a profile
        profile = _server_profiles[default_profile]

      client = _client_connections['__default__'] = impl(**profile)
      return client

    else:  # pragma: no cover

      # mock adapter testing
      import fakeredis

      if not _mock_redis:
        _mock_redis = fakeredis.FakeStrictRedis()
      return _mock_redis

  @classmethod
  def execute(cls, operation, kind, *args, **kwargs):

    """ Acquire a channel and execute an operation, optionally buffering the
        command.

        :param operation: Operation name to execute (from
          :py:attr:`RedisAdapter.Operations`).

        :param kind: String :py:class:`model.Model` kind to acquire the channel
          for.

        :param args: Positional arguments to pass to the low-level operation
          selected.

        :param kwargs: Keyword arguments to pass to the low-level operation
          selected.

        :returns: Result of the selected low-level operation. """

    # defer to pipeline or resolve channel for kind
    target = kwargs.get('target', None)
    if target is None: target = cls.channel(kind)

    if operation == cls.Operations.DELETE:
      # special case: `delete` instead of `del` (because it's a keyword)
      operation = 'DELETE'

    if 'target' in kwargs: del kwargs['target']

    try:
      if isinstance(operation, tuple):  # pragma: no cover
        # (CLIENT, KILL) => "CLIENT KILL"
        operation = '_'.join(map(unicode, operation))
      if isinstance(target, (_redis_client.client.Pipeline,
                             _redis_client.client.StrictPipeline)) or (
              fakeredis and isinstance(target, fakeredis.FakePipeline)):
        getattr(target, operation.lower())(*args, **kwargs)
        return target
      if operation == cls.Operations.HASH_SET:  # pragma: no cover
        r = getattr(target, operation.lower())(*args, **kwargs)
        if r in (0, 1):
          # count 0 and 1 as success, as it indicates an overwrite,
          # not a failure
          return 1
      return getattr(target, operation.lower())(*args, **kwargs)
    except Exception:  # pragma: no cover
      raise

  @classmethod
  def inflate(cls, result):
    """ Small closure that can inflate (and potentially decompress) a resulting
        object for a given storage mode.

        :param result: ``basestring`` result from raw Redis storage.

        :returns: Inflated entity, if applicable. """

    if not isinstance(result, basestring):  # pragma: no cover
      return result  # accounts for dict-like storage

    # account for none, optionally decompress
    if cls.EngineConfig.compression:  # pragma: no cover
      try:
        result = cls.compressor.decompress(result)
      except:
        pass  # maybe entity is uncompressed? will fail during deserialization

    # deserialize structures
    return cls.serializer.loads(result)

  @classmethod
  def get(cls, key, pipeline=None, _entity=None):

    """ Retrieve an entity by Key from Redis.

        :param key: Target :py:class:`model.Key` to retrieve from storage.

        :param pipeline: Redis pipeline to enqueue the resulting commands
          in, rather than directly executing them. Defaults to ``None``. If a
          pipeline is passed, it will be returned in lieu of the pending
          result.

        :param _entity: Entity to inflate, if we already have one.

        :returns: The deserialized and decompressed entity associated with the
          target ``key``. """

    from canteen import model

    if key:
      encoded, flattened = key

      # @TODO(sgammon): access to structured keys in adapters
      joined, _ = model.Key.from_urlsafe(encoded).flatten(True)

      ## toplevel_blob
      if cls.EngineConfig.mode == RedisMode.toplevel_blob:

        # execute query
        result = _entity or (
          cls.execute(cls.Operations.GET, flattened[1],
                        encoded,
                        target=pipeline))

      ## hashkind_blob
      elif cls.EngineConfig.mode == RedisMode.hashkind_blob:

        # generate kinded key and trim tail
        j_kinded, f_kinded = model.Key(flattened[1]).flatten(True)
        tail = joined.replace(j_kinded, '')

        result = _entity or (
          cls.execute(*(
            cls.Operations.HASH_GET,
            flattened[1],
            cls.encode_key(j_kinded, f_kinded),
            cls.encode_key(tail, flattened)), target=pipeline))

      ## hashkey_blob
      elif cls.EngineConfig.mode == RedisMode.hashkey_blob:

        # build key and extract group
        desired_key = model.Key.from_raw(joined)
        root = (ancestor for ancestor in desired_key.ancestry).next()
        tail = (
          desired_key.flatten(True)[0].replace(root.flatten(True)[0], '') or (
            '__root__'))

        result = _entity or (
          cls.execute(*(
            cls.Operations.HASH_GET,
            flattened[1],
            cls.encode_key(*root.flatten(True)),
            cls.encode_key(tail, flattened)), target=pipeline))

      ## hashkey_hash
      elif cls.EngineConfig.mode == RedisMode.hashkey_hash:  # pragma: no cover

        raise NotImplementedError('Redis mode not implemented: "hashkey_hash".')

      else:  # pragma: no cover
        raise NotImplementedError("Unknown storage mode: '%s'." % (
                                                        cls.EngineConfig.mode))

    else:  # pragma: no cover
      result = _entity

    # @TODO: different storage internal modes
    if isinstance(result, basestring):
      return cls.inflate(result)
    return result

  def _put(self, entity, **kwargs):  # pragma: no cover

    """ Overrides low-level ``put`` process to enable pipelining of object
        writes with their indexes.

        :param entity: Entity :py:class:`model.Model` to persist.

        :returns: Resulting :py:class:`model.Key` from write operation. """

    _indexed_properties = self._pluck_indexed(entity)

    # reuse pipeline passed, if any
    if 'pipeline' in kwargs:
      pipeline = kwargs['pipeline']
      del kwargs['pipeline']
    else:
      pipeline = (
        self.channel(entity.kind()).pipeline(transaction=True))

    # provision ID early if there is none
    if not entity.key.id:
      entity.key.id = self.allocate_ids(entity.__keyclass__, entity.kind())

    with pipeline as pipe:

      # delegate write up the chain
      written_key = super(IndexedModelAdapter, self)._put(entity,
                                                        pipeline=pipe, **kwargs)

      # proxy to `generate_indexes` and write indexes
      origin, meta, property_map, graph = (
        self.generate_indexes(entity.key, entity, _indexed_properties))

      self.write_indexes((origin, meta, property_map), graph,
                          pipeline=pipe, **kwargs)

      # collapse pipelines
      pipe.execute()
      return written_key  # delegate up the chain for entity write

  @classmethod
  def put(cls, key, entity, model, pipeline=None):

    """ Persist an entity to storage in Redis.

        :param key: New (and potentially empty) :py:class:`model.Key` for
          ``entity``. Must be assigned an ``ID`` by the driver through
          :py:meth:`RedisAdapter.allocate_ids` in the case of an empty
          (non-deterministic) :py:class:`model.Key`.

        :param entity: Object entity :py:class:`model.Model` to persist in
          ``Redis``.

        :param model: Schema :py:class:`model.Model` associated with the target
          ``entity`` being persisted.

        :param pipeline: Existing pipeline of queued commands to append to, if
          applicable.

        :returns: Result of the lower-level write operation. """

    from canteen import model as _model

    # reduce entity to dictionary
    serialized = entity if isinstance(entity, dict) else (
      entity.to_dict(convert_datetime=False,
                     convert_keys=True,
                     convert_models=True))
    joined, flattened = key

    # clean key types
    _cleaned = {}
    for k, v in serialized.iteritems():
      prop = getattr(model, k)
      if isinstance(v, (datetime.date, datetime.time, datetime.datetime)):
        _cleaned[k] = v.isoformat()  # pragma: no cover
      else:
        _cleaned[k] = v

    # serialize + optionally compress
    serialized = cls.serializer.dumps(_cleaned)
    if cls.EngineConfig.compression:  # pragma: no cover
      compressed = cls.compressor.compress(serialized)

      if len(compressed) < len(serialized):
        # we saved space, store it compressed and it should uncompress on `get`
        serialized = compressed

    # toplevel_blob
    if cls.EngineConfig.mode == RedisMode.toplevel_blob:

      # delegate to redis client
      if cls.execute(*(
          cls.Operations.SET,
          flattened[1],
          joined,
          serialized), target=pipeline):
        entity._set_persisted(True)
        return entity.key
      else:  # pragma: no cover
        raise RuntimeError('Failed to write entity "%s" to key "%s".' % (
          str(entity), str(key) or '<none>'))

    ## need a serialized blob...

    ## hashkind_blob
    if cls.EngineConfig.mode == RedisMode.hashkind_blob:

      # generate kinded key and trim tail
      kinded = _model.Key(flattened[1]).flatten(True)
      tail = entity.key.flatten(True)[0].replace(kinded[0], '')

      # delegate to redis client
      if cls.execute(*(
        cls.Operations.HASH_SET,
        flattened[1],
        cls.encode_key(*kinded),
        cls.encode_key(tail, flattened),
        serialized), target=pipeline):
        entity._set_persisted(True)
        return entity.key
      else:  # pragma: no cover
        raise RuntimeError('Failed to write entity "%s" to key "%s".' % (
          str(entity), str(key) or '<none>'))

    ## hashkey_blob
    elif cls.EngineConfig.mode == RedisMode.hashkey_blob:

      # find entity group key
      root = (ancestor for ancestor in entity.key.ancestry).next().flatten(True)
      tail = entity.key.flatten(True)[0].replace(root[0], '') or '__root__'

      if cls.execute(*(
        cls.Operations.HASH_SET,
        flattened[1],
        cls.encode_key(*root),
        cls.encode_key(tail, flattened),
        serialized), target=pipeline):
        entity._set_persisted(True)
        return entity.key
      else:  # pragma: no cover
        raise RuntimeError('Failed to write entity "%s" to key "%s".' % (
          str(entity), str(key) or '<none>'))

    ## hashkey_hash
    elif cls.EngineConfig.mode == RedisMode.hashkey_hash:  # pragma: no cover

      raise NotImplementedError('Redis mode not implemented: "hashkey_hash".')

    raise NotImplementedError("Unknown storage mode: '%s'." % (
                              cls.EngineConfig.mode))  # pragma: no cover

    # @TODO: different storage internal modes

  # @TODO(sgammon): testing for ability to delete entities

  @classmethod
  def delete(cls, key, pipeline=None):

    """ Delete an entity by Key from Redis.

        :param key: Target :py:class:`model.Key`, whose associated
          :py:class:`model.Model` is being deleted.

        :returns: The result of the low-level delete operation. """

    from canteen import model

    # @TODO(sgammon): access to structured keys in adapters

    encoded, flattened = key
    try:
      joined, _ = model.Key.from_urlsafe(encoded).flatten(True)
    except TypeError:  # pragma: no cover
      joined = encoded
      encoded = cls.encode_key((joined, flattened))


    if cls.EngineConfig.mode == RedisMode.toplevel_blob:

      # delegate to redis client with encoded key
      return cls.execute(*(
        cls.Operations.DELETE,
        flattened[1],
        cls.encode_key(joined, flattened)), target=pipeline)

    elif cls.EngineConfig.mode == RedisMode.hashkind_blob:

      # generate kinded key and trim tail
      kinded = model.Key(flattened[1]).flatten(True)
      tail = joined.replace(kinded[0], '')

      # delegate to redis client
      return cls.execute(*(
        cls.Operations.HASH_DELETE,
        flattened[1],
        cls.encode_key(*kinded),
        cls.encode_key(tail, flattened)), target=pipeline)

    elif cls.EngineConfig.mode == RedisMode.hashkey_blob:

      # build key and extract group
      desired_key = model.Key.from_raw(joined)
      root = (ancestor for ancestor in desired_key.ancestry).next()
      tail = (
        desired_key.flatten(True)[0].replace(root.flatten(True)[0], '') or (
          '__root__'))

      return cls.execute(*(
        cls.Operations.HASH_DELETE,
          flattened[1],
          cls.encode_key(*root.flatten(True)),
          cls.encode_key(tail, flattened)), target=pipeline)

    elif cls.EngineConfig.mode == RedisMode.hashkey_hash:  # pragma: no cover

      raise NotImplementedError('Redis mode not implemented: "hashkey_hash".')

    raise NotImplementedError("Unknown storage mode: '%s'." % (
                              cls.EngineConfig.mode))  # pragma: no cover

  @classmethod
  def allocate_ids(cls, key_class, kind, count=1, pipeline=None):

    """ Allocate new :py:class:`model.Key` IDs up to ``count``. Allocated IDs
        are guaranteed not to be provisioned or otherwise used by the underlying
        persistence engine, and thus can be used for uniquely identifying
        non-deterministic data.

        :param key_class: Descendent of :py:class:`model.Key` to allocate IDs
          for.

        :param kind: String :py:class:`model.Model` kind name.

        :param count: The number of IDs to generate, which **must** be greater
          than 1. Defaults to ``1``.

        :raises ValueError: In the case the ``count`` is less than ``1``.

        :returns: If **only one** ID is requested, an **integer ID** suitable
          for use in a :py:class:`model.Key` directly. If **more than one** ID
          is requested, a **generator** is returned, which ``yields`` a set of
          provisioned integer IDs, each suitable for use in a
          :py:class:`model.Key` directly. """

    if not count:  # pragma: no cover
      raise ValueError("Cannot allocate less than 1 ID's.")

    # generate kinded key to resolve ID pointer
    kinded_key = key_class(kind)
    joined, flattened = kinded_key.flatten(True)

    if cls.EngineConfig.mode == RedisMode.toplevel_blob:
      key_root_id = cls._magic_separator.join([
        cls._meta_prefix, cls.encode_key(joined, flattened)])

      # increment by the amount desired
      value = cls.execute(*(
        cls.Operations.HASH_INCREMENT,
        kinded_key.kind,
        key_root_id,
        cls._id_prefix,
        count), target=pipeline)

    elif cls.EngineConfig.mode in (
      RedisMode.hashkind_blob,
      RedisMode.hashkey_blob):

      # store auto-increment for kind in kind's own hash at special field
      # ends up as `__meta__::id` or so
      tail = cls._magic_separator.join([cls._meta_prefix, 'id'])

      # delegate to redis client
      value = cls.execute(*(
        cls.Operations.HASH_INCREMENT,
        flattened[1],
        cls.encode_key(joined, flattened),
        cls.encode_key(tail, flattened),
        count), target=pipeline)

    elif cls.EngineConfig.mode == RedisMode.hashkey_hash:  # pragma: no cover

      raise NotImplementedError('Redis mode not implemented: "hashkey_hash".')

    else:  # pragma: no cover

      raise NotImplementedError("Unknown storage mode: '%s'." % (
                                cls.EngineConfig.mode))

    if count > 1:  # pragma: no cover
      def _generate_range():

        """ Generate a range of requested ID's.

            :yields: Each item in a set of provisioned integer IDs,
              suitable for use in a :py:class:`model.Key`. """

        bottom_range = (value - count)
        for i in xrange(bottom_range, value):
          yield i

      return _generate_range
    return value

  @classmethod
  def encode_key(cls, joined, flattened=None):  # pragma: no cover

    """ Encode a Key for storage in ``Redis``. Since we don't need to
        do anything fancy, just delegate this to the abstract (default)
        encoder, which is ``base64``.

        If :py:attr:`RedisEngine.EngineConfig.encoding` is disabled, this
        simply returns the ``joined`` :py:class:`model.Key` (for reference,
        see :py:meth:`model.Key.flatten`).

        :param joined: String-joined :py:class:`model.Key`.

        :param flattened: Tupled ("raw format") :py:class:`model.Key`.

        :returns: In the case that ``encoding`` is *on*, the encoded string
          :py:class:`model.Key`, suitable for storage in ``Redis``. Otherwise
          (``encoding`` is *off*), the cleartext ``joined`` key. """

    from canteen import model

    if isinstance(joined, model.Key): return joined.urlsafe()
    if cls.EngineConfig.encoding: return abstract._encoder(joined)
    return joined

  @classmethod
  def write_indexes(cls, w, g, pipeline=None, execute=True):

    """ Write a batch of index updates generated earlier via
        :py:meth:`RedisAdapter.generate_indexes`.

        :param w: Batch of writes to commit to ``Redis``.
        :param g: Batch of graph writes to commit to ``Redis``.

        :param pipeline: Current active pipeline of ``Redis`` commands to
          collapse and execute. Defaults to ``None``, which either returns the
          queued commands (if ``execute`` is ``False``) or a list of results
          from those commands (if ``execute`` is ``True``). Always returned
          if not ``None``, so that other calls can be chained.

        :param execute: Whether we should actually execute the writes, or just
          plan it and send it back. ``bool``, defaults to ``True``, which *does*
          execute the writes.

        :raises RuntimeError:

        :returns: ``pipeline`` if ``pipeline`` was not ``None``, or a ``tuple``
          of operation results if ``execute`` was ``True`` and ``pipeline`` was
          ``None``, or a tuple of plans that *would* be executed if ``pipeline``
           is ``None`` and ``execute`` is ``False``. """

    from .. import Vertex, Edge, VertexKey, EdgeKey

    origin, meta, property_map = w

    results, indexer_calls = [], []

    # resolve target (perhaps a pipeline?)
    if pipeline:  # pragma: no cover
      target = pipeline
    else:  # pragma: no cover
      target = cls.channel(cls._meta_prefix)

    handler, ekey_encoder, vkey_encoder = (
      cls.Operations.SET_ADD,
      cls._index_basetypes[EdgeKey],
      cls._index_basetypes[VertexKey])

    # write graph indexes
    for bundle in g:  # pragma: no cover

      bundle_args = []

      # one- or two-element tuples are simple indexes (and always edges)
      if 0 < len(bundle) < 3:
        bundle_args.append(Edge)
        bundle_args.append(cls._magic_separator.join(bundle))
        bundle_args.append(origin)

      # three-element tuples are encoded key indexes
      elif len(bundle) == 3:

        # encode key components
        _components = []
        for _item in bundle:
          if isinstance(_item, EdgeKey):
            _components.append(ekey_encoder(_item)[1])
          elif isinstance(_item, VertexKey):
            _components.append(vkey_encoder(_item)[1])
          elif isinstance(_item, basestring):
            _components.append(_item)
          else:  # pragma: no cover
            raise ValueError('Got invalid value in graph index bundle'
                             ' that was neither a `basestring` or `VertexKey`'
                             ' or `EdgeKey`. Instead, got: "%s"'
                             ' of type "%s".' % (repr(_item), type(_item)))

        # slice key and value into args
        gbase, gtoken, gtarget = tuple(_components)

        bundle_args.append(Vertex)
        bundle_args.append(cls._magic_separator.join((cls._graph_prefix,
                                                      gbase,
                                                      gtoken)))
        bundle_args.append(gtarget)

      # invalid indexer bundle
      else:  # pragma: no cover
        raise RuntimeError('Invalid graph index bundle: "%s".' % bundle)

      # build index key
      indexer_calls.append((handler, tuple(bundle_args), {'target': target}))

    # write meta and property indexes
    for btype, bundle in (('meta', meta), ('property', property_map)):

      # add meta indexes first
      for element in bundle:  # pragma: no cover

        # provision args, kwargs, and hash components containers
        args, kwargs, hash_c = [], {}, []

        if btype == 'meta':
          # meta indexes are always unicode
          converter, write = unicode, element
        else:
          # property indexes come through with a type converter
          converter, write = element

          # basestring is not allowed to be instantiated
          if converter is basestring:
            converter = cls.serializer.dumps

        ## Unpack index write
        if len(write) > 3:  # qualified value-key-mapping index

          # extract write, inflate
          index, path, value = write[0], write[1:-1], write[-1]
          hash_c.append(index)
          hash_c.append(cls._path_separator.join(path))
          hash_value = True  # add hashed value later

        elif len(write) == 3:  # value-key-mapping index

          # extract write, inflate
          index, path, value = write
          hash_c.append(index)
          hash_c.append(cls._path_separator.join(path))
          hash_value = True  # add hashed value later

        elif len(write) == 2:  # it's a qualified key index

          # extract write, inflate
          index, value = write
          path = None
          hash_c.append(index)
          hash_value = True  # do not add hashed value later

        elif len(write) == 1:  # it's a simple key index

          # extract index, add value
          index, path, value = write[0], None, origin
          hash_c.append(index)
          hash_value = False  # do not add hashed value later

        else:
          raise ValueError('Invalid index write bundle: "%s".' % write)

        # resolve datatype of index
        if not isinstance(value, bool) and isinstance(value, _SERIES_BASETYPES):

          if converter is None:  # pragma: no cover
            raise RuntimeError('Illegal non-string value passed in'
                               ' for a `meta` index value.'
                               ' Write bundle: "%s".' % write)

          # time-based or number-like values are stored in sorted sets
          handler = cls.Operations.SORTED_ADD
          sanitized_value = converter(value)

          if isinstance(sanitized_value, tuple):
            magic_symbol, sanitized_value = sanitized_value
            hash_c.append(unicode(magic_symbol))
            args.append(sanitized_value)  # add calculated score to args
          else:
            args.append(sanitized_value)
          args.append(origin)  # add key to args

        else:

          # @TODO(sgammon): this can cause silent issues

          # everything else is stored unsorted
          handler = cls.Operations.SET_ADD

          if converter and value is not None:
            sanitized_value = converter(value)
          else:
            sanitized_value = value

          if hash_value:
            hash_c.append(sanitized_value)
          args.append(origin)

        # build index key
        indexer_calls.append((handler, tuple([
          None, cls._magic_separator.join(map(str, hash_c))] + args), {
            'target': target}))

    if execute:  # pragma: no cover
      for handler, hargs, hkwargs in indexer_calls:
        results.append(cls.execute(handler, *hargs, **hkwargs))

      if pipeline:
        return pipeline
      return results
    return indexer_calls  # pragma: no cover

  @classmethod
  def clean_indexes(cls, writes, pipeline=None):  # pragma: no cover

    """ Clean indexes and index entries matching a particular
        :py:class:`model.Key`, and generated via the adapter method
        :py:meth:`RedisAdapter.generate_indexes`.

      :param writes: Writes to clean up.

      :returns: ``None``. """

    return  # not currently implemented

  @classmethod
  def execute_query(cls, kind, spec, options, **kwargs):  # pragma: no cover

    """ Execute a :py:class:`model.Query` across one (or multiple) indexed
        properties.

        :param kind: Kind name (``str``) for which we are querying across, or
          ``None`` if this is a ``kindless`` query.

        :param spec: Tupled pair of ``filter`` and ``sort`` directives to apply
          to this :py:class:`Query` (:py:class:`query.Filter` and
          :py:class:`query.Sort` and their descendents, respectively), like
          ``(<filters>, <sorts>)``.

        :param options: Object descendent from, or directly instantiated as
          :py:class:`QueryOptions`, specifying options for the execution of
          this :py:class:`Query`.

        :param **kwargs: Low-level options for handling this query, such as
          ``pipeline`` (for pipelining support) and ``execute`` (to trigger a
          buffer flush for a generated or constructed pipeline or operation
          buffer).

        :returns: Iterable (``list``) of matching :py:class:`model.Key` yielded
          by execution of the current :py:class:`Query`. Returns
          empty iterable (``[]``) in the case that no results could be
          found. """

    # @TODO(sgammon): desparately needs rewriting. absolute utter plebbery.

    from canteen import model
    from canteen.model import query

    if not kind:
      # @TODO(sgammon): implement kindless queries
      raise NotImplementedError('Kindless queries are not yet implemented'
                                ' in `Redis`.')

    # extract filter and sort directives and build ancestor
    filters, sorts = spec
    _data_frame = []  # allocate results window

    # calculate ancestry parent
    ancestry_parent = None
    if isinstance(options.ancestor, basestring):
      ancestry_parent = model.Key.from_urlsafe(options.ancestor)
    elif isinstance(options.ancestor, model.Key):
      ancestry_parent = options.ancestor
    elif isinstance(options.ancestor, model.Model):
      ancestry_parent = options.ancestor.key

    _and_filters, _or_filters = [], []

    if filters:
      kinded_key = model.Key(kind, parent=ancestry_parent)

      ## HUGE HACK: use generate_indexes to make index names.
      # fix this plz

      _filters, _filter_i_lookup = {}, set()
      for _f in filters:

        # handle graph-based edge/neighbor filters first
        if isinstance(_f, query.EdgeFilter):

          # extract property values
          _data = _f.value.data if (
            isinstance(_f.value, model.Model._PropertyValue)) else _f.value

          # graph prefix and search key
          _index_key = [
            cls._graph_prefix,
            cls._index_basetypes[model.VertexKey](_data)[1]]

          # edge queries
          if _f.kind is _f.EDGES:

            # undirected edge queries
            if _f.tails is None:
              _index_key.append(cls._peers_token)

            else:
              # directed edge queries
              _index_key.append(cls._out_token if _f.tails else cls._in_token)

          # neighbor queries
          elif _f.kind is _f.NEIGHBORS:
            _index_key.append(cls._neighbors_token)

          else:  # pragma: no cover
            raise RuntimeError('Invalid `EdgeFilter` kind: "%s".' % _f.kind)

          # check for uniqueness and add to queued, unsorted filters
          _filter_key = ('S', cls._magic_separator.join(_index_key))
          if _filter_key not in _filter_i_lookup:
            _filter_i_lookup.add(_filter_key)
            _filters[_filter_key] = [(_f.operator, _f.value, _f.chain)]

        # then handle property/meta filters, etc
        else:
          origin, meta, property_map, graph_indexes = (
            cls.generate_indexes(*(
              kinded_key, None, {_f.target.name: (_f.target, _f.value.data)})))

          for operation, index, config in cls.write_indexes(
              (origin, [], property_map), graph_indexes, execute=False):

            if operation == cls.Operations.SORTED_ADD:
              _flag, _index_key, value = 'Z', index[1], index[2]
            else:
              _flag, _index_key, value = 'S', index[1], index[2]

          if (_flag, _index_key) not in _filter_i_lookup:
            _filters[(_flag, _index_key)] = []
            _filter_i_lookup.add((_flag, _index_key))

          _filters[(_flag, _index_key)].append((_f.operator, value, _f.chain))

      # process sorted sets first: leads to lower cardinality
      sorted_indexes = dict([
        (index, _filters[index]) for index in (
          filter(lambda x: x[0] == 'Z', _filters.iterkeys()))])

      unsorted_indexes = dict([
        (index, _filters[index]) for index in (
          filter(lambda x: x[0] == 'S', _filters.iterkeys()))])

      if sorted_indexes:

        for prop, _directives in sorted_indexes.iteritems():

          _operator, _value, chain = _directives[0]

          if chain:
            for subquery in chain:
              if subquery.sub_operator is query.AND:
                _and_filters.append(subquery)
              if subquery.sub_operator is query.OR:
                _or_filters.append(subquery)

          # double-filters
          if len(_filters[prop]) == 2:

            # extract info
            _operators = [operator for operator, value in _directives]
            _values = [value for operator, value in _directives]
            _, prop = prop

            # special case: maybe we can do a sorted range request
            if (query.GREATER_THAN in _operators) or (
                    query.GREATER_THAN_EQUAL_TO in _operators):

              if (query.LESS_THAN in _operators) or (
                    query.LESS_THAN_EQUAL_TO in _operators):

                # range value query over sorted index
                greater, lesser = max(_values), min(_values)
                _data_frame.append(cls.execute(*(
                  cls.Operations.SORTED_RANGE_BY_SCORE,
                  None,
                  prop,
                  lesser,
                  greater,
                  options.offset,
                  options.limit)))

                continue

          if len(_filters[prop]) == 1:  # single-filters

            # extract info
            _operator = [operator for operator, value in _directives][0]
            _value = float([value for operator, value in _directives][0])
            _, prop = prop

            if _operator is query.EQUALS:

              # static value query over sorted index
              _data_frame.append(cls.execute(*(
                cls.Operations.SORTED_RANGE_BY_SCORE,
                None,
                prop,
                _value,
                _value,
                options.offset,
                options.limit)))

              continue

          ## @TODO(sgammon): build this query branch
          raise RuntimeError("Specified query is not yet supported.")

      if unsorted_indexes:

        _intersections = set()
        for prop, _directives in unsorted_indexes.iteritems():

          _operator, _value, chain = _directives[0]

          if chain:
            for subquery in chain:
              if subquery.sub_operator is query.AND:
                _and_filters.append(subquery)
              if subquery.sub_operator is query.OR:
                _or_filters.append(subquery)

          _flag, index = prop
          _intersections.add(index)

        # special case: only one unsorted set - pull content instead
        # of an intersection merge
        if _intersections and len(_intersections) == 1:
          _data_frame.append(cls.execute(*(
            cls.Operations.SET_MEMBERS,
            None,
            _intersections.pop())))

        # more than one intersection: do an `SINTER` call instead of `SMEMBERS`
        if _intersections and len(_intersections) > 1:
          _data_frame.append(cls.execute(*(
            cls.Operations.SET_INTERSECT,
            None,
            _intersections)))

      if _data_frame:  # there were results, start merging
        _result_window = set()
        for frame in _data_frame:

          if not len(_result_window):
            _result_window = set(frame)
            continue  # initial frame: fill background

          _result_window = (_result_window & set(frame))
        matching_keys = [k for k in _result_window]
      else:
        matching_keys = []

    else:
      if not sorts:

        # pull all entities for this kind
        kind_index_name = None
        encoded, indexes, property_map = (
          cls.generate_indexes(model.Key(kind, 0), None, None))
        for bundle in indexes:
          if len(bundle) == 2:
            metakey, metaprefix = bundle
            if metakey == cls._kind_prefix:
              matching_keys = cls.execute(*(cls.Operations.SET_MEMBERS,
                                            kind.kind(),
                                            cls._magic_separator.join(bundle)))

          else:
            continue  # only interested in meta indexes for kinds

        # if we're doing keys only, we're done
        if options.keys_only and not (_and_filters or _or_filters):

          results = []
          for k in matching_keys:
            # @TODO(sgammon): unambiguous key classes
            vanilla = model.Key.from_urlsafe(k, _persisted=True)
            if vanilla.kind != _base_kind.kind():
              _base_kind = cls.registry.get(vanilla.kind, _base_kind)
              results.append(
                _base_kind.__keyclass__.from_urlsafe(k, _persisted=True))
            else:
              results.append(vanilla)
          return results

    result_entities = []  # otherwise, build entities and return

    # fetch keys
    with cls.channel(kind.kind()).pipeline(transaction=False) as pipeline:

      # fill pipeline
      _seen_results, _queued = 0, collections.deque()
      for key in matching_keys:
        if (not (_and_filters or _or_filters)) and (
                0 < options.limit <= _seen_results):
          break
        _seen_results += 1

        decoded_k, _base_kind = (
          model.Key.from_urlsafe(key, _persisted=True), None)
        if not decoded_k.kind == kind.kind():
          _base_kind = cls.registry.get(kind.kind())
        if decoded_k.kind == kind.kind() or not _base_kind:
          _base_kind = kind

        if not _base_kind:  # pragma: no cover
          raise TypeError('Unknown model kind: "%s".' % decoded_k.kind)

        # @TODO(sgammon): make vertex/edge keys unambiguous
        decoded_k = _base_kind.__keyclass__.from_urlsafe(key)

        # queue fetch of key
        _queued.append(decoded_k)

        joined, flattened = decoded_k.flatten(True)
        pipeline = cls.get((cls.encode_key(joined, flattened), flattened),
                           pipeline=pipeline)

        if (options.limit > 0) and _seen_results >= options.limit:
          break  # break early to avoid pulling junk entities

      _blob_results = pipeline.execute()

      # execute pipeline, zip keys and build results
      for key, entity in zip(_queued, _blob_results):
        if not entity: continue  # skip entities that couldn't be found

        # decode raw entity + attach key
        decoded = cls.get(key=None, _entity=entity)

        decoded['key'] = key

        if _and_filters or _or_filters:
          if _and_filters and not all((
                  (_filter.match(decoded) for _filter in _and_filters))):
            continue  # doesn't match one of the filters

          if _or_filters and not any((
                  (_filter.match(decoded) for _filter in _or_filters))):
            continue  # doesn't match any of the `or` filters

        # if key kind doesn't match base, find suitable impl class
        if key.kind != _base_kind.kind():
          _base_kind = cls.registry.get(key.kind, _base_kind)
          decoded['key'] = (
            _base_kind.__keyclass__.from_urlsafe(decoded['key'].urlsafe()))

        # inflate and get ready to return
        _inflated = _base_kind(_persisted=True, **decoded)
        result_entities.append(_inflated.key if options.keys_only else (
                               _inflated))
    return result_entities



# -*- coding: utf-8 -*-

"""

  core injection tests
  ~~~~~~~~~~~~~~~~~~~~

  tests the injection core, which is the "frontend" to the DI system.
  "compound" meta-objects, defined by the module tested by this code,
  can make use of the attribute injection pool locally via MRO.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""



# -*- coding: utf-8 -*-

"""

  model meta tests
  ~~~~~~~~~~~~~~~~

  tests metacomponents of canteen's model layer.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import inspect

# canteen model API
from canteen import model
from canteen.model import adapter
from canteen.model import MetaFactory

# canteen tests
from canteen.test import FrameworkTest


## MetaFactoryTests
class MetaFactoryTests(FrameworkTest):

  """ Tests `model.MetaFactory`. """

  def test_abstract_factory(self):

    """ Test that `MetaFactory` is only usable abstractly. """

    # constructing metafactory should raise an ABC exception
    self.assertTrue(inspect.isabstract(MetaFactory))
    with self.assertRaises(NotImplementedError):
      MetaFactory()

  def test_abstract_enforcement(self):

    """ Test abstraction enforcement on `MetaFactory` """

    class InsolentClass(MetaFactory):

      """ Look at me! I extend without implementing. The nerve! """

      # intentionally not defined: def classmethod(initialize())
      pass

    with self.assertRaises(TypeError):
      InsolentClass(*(InsolentClass.__name__,
                      (MetaFactory, type),
                      dict([
                        (k, v) for k, v in InsolentClass.__dict__.items()])))

  def test_resolve_adapters(self):

    """ Test that `MetaFactory` resolves adapters correctly """

    # test that resolve exists
    self.assertTrue(inspect.ismethod(MetaFactory.resolve))
    self.assertIsInstance(MetaFactory.resolve(*(model.Model.__name__,
                                                model.Model.__bases__,
                                                model.Model.__dict__,
                                                False)), tuple)

    self.assertIsInstance(MetaFactory.resolve(*(model.Model.__name__,
                                                model.Model.__bases__,
                                                model.Model.__dict__,
                                                True)), adapter.ModelAdapter)



# -*- coding: utf-8 -*-

"""

  core hooks
  ~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import inspect

# core runtime
from . import runtime


class HookResponder(object):

  """ Provides an object that can hook into named points in runtime execution
      flow. Context is provided as keyword arguments and may be subscribed to by
      item name.

      A full list of hook points or context items is not yet available. """

  # @TODO(sgammon): make list of hook points and context items

  __slots__ = (
    '__func__',  # inner function
    '__wrap__',  # hook wrapper
    '__hooks__',  # event names to fire on
    '__argspec__',  # argspec (explicit or implied)
    '__binding__')  # binding to carry through if wrap is a bind

  def __init__(self, *events, **kwargs):

    """ Initialize this ``HookResponder``.

        :param *events: Iterable of event names to subscribe to.
        :param **kwargs: Configuration, notably ``wrap`` (which can be used to
          re-wrap the target callable). """

    self.__hooks__, self.__argspec__, self.__wrap__ = (
      frozenset(events),  # events to fire on
      Context(
        kwargs.get('context'),  # explicit argspec
        kwargs.get('rollup', False),  # kwargs flag
        kwargs.get('notify', False),  # event notify
      ) if kwargs else None,
      kwargs.get('wrap'))  # function to wrap the hook in, if any

  def __register__(self, context):

    """ Register this ``HookResponder`` with the currently-active runtime, which
        will make it available when hooks are due to be executed.

        :param context: Requested context to register alongside this
          ``HookResponder``.

        :returns: Nothing. """

    for i in self.__hooks__:  # add hook for each event name
      runtime.Runtime.add_hook(i, (context, self))

  def __call__(self, *args, **kwargs):

    """ Execute this local ``HookResponder``, which will dispatch the underlying
        hook target, passing along any arguments and keyword arguments.

        :param **args: Positional arguments to pass to the target callable.
        :param **kwargs: Keyword arguments to pass to the target callable.

        :returns: Whatever the target callable returns. """

    from ..util import decorators

    if not hasattr(self, '__func__') or not getattr(self, '__func__'):
      # if there's no explicit argspec, inspect
      hook = args[0]
      if not self.__argspec__:
        _hook_i = inspect.getargspec(hook)
        self.__argspec__ = Context([i for i in _hook_i.args if i not in (
          'self', 'cls')], _hook_i.keywords is not None)

      # carry through DI bindings
      self.__binding__ = hook.__binding__ if (
        isinstance(self.__wrap__, decorators.bind)) else None

      # noinspection PyCallingNonCallable
      def run_hook(*args, **kwargs):

        """ Execute the local hook according to the configuration held by the
            encapsulating ``HookResponder``.

            :param *args: Positional arguments to pass to the hook.
            :param **kwargs: Keyword arguments to pass to the hook.

            :returns: Whatever the hook returns. """

        return self.__argspec__((
          self.__wrap__(hook) if self.__wrap__ else hook))(*args, **kwargs)
      return setattr(self, '__func__', run_hook) or self  # mount run_hook
    return self.__func__(*args, **kwargs)


class Context(object):

  """ Object that contains context for a given ``HookResponder`` instance. Holds
      hook kwargs and args for target execution. """

  __slots__ = (
    '__requested__',  # requested args
    '__rollup__',  # acceptance of kwargs
    '__notify__')  # requested hook name

  def __init__(self, requested, rollup=True, notify=False):

    """ Initialize this ``HookResponder`` ``Context`` object.

        :param requested: Context items that are explicitly requested to be
          provided at runtime.

        :param rollup: ``Bool``, indicating support in the target callable for
          accepting a ``**kwargs``-style rolled-up set of context items,
          including extra (unrequested) context items.

        :param notify: ``Bool``, indicating that the target expects the event
          name for which it is being called to be inserted as the first
          positional argument. """

    self.__requested__, self.__rollup__, self.__notify__ = (
      requested, rollup, notify)

  def __call__(self, func):

    """ Pair this ``Context`` with the target ``func`` and execute using the
        locally-attached ``requested`` args, potentially using ``rollup``.

        :param func: Target function to wrap with a closure to properly call it
          with the provided context.

        :raises RuntimeError: In the inner hook closure, if a case arises where
          a target requests a context item that is not available.

        :returns: ``with_context`` inner closure, that applies stored context to
          the target ``func`` when dispatched. """

    def with_context(*args, **context):

      """ Closure returned to execute args and context items with a provided
          target ``func``, usually the backing to a ``HookResponder``. Arguments
          are passed through to the target callable.

          Accepts positional and keyword arguments on behalf of the wrapped
          ``func``.

          :param args: Positional arguments to pass to target hook responder.

          :param context: Keyword arguments (considered as "context" in this,
            well, context) to pass to the target hook responder.

          :raises RuntimeError: If a case is encountered where a hook function
            requests a context item that is not yet available in the runtime
            execution flow.

          :returns: Result of calling the target ``func`` with applied ``args``
            and ``context``. """

      # extract hookname from args (always 1st param)
      hookname, args = args[0], args[1:]

      # calculate materialized args
      _args, _kwargs = [], {}
      if self.__requested__:
        for prop in self.__requested__:
          if prop not in context:
            raise RuntimeError('Cannot satisfy request for context entry `%s`'
                               ' in hook `%s` for event point `%s`.' % (
                                prop,
                                (func if not (
                                  isinstance(func, (classmethod, staticmethod)))
                                  else func.__func__.__name__), hookname))
          _args.append(context[prop])

      # honor kwargs
      if self.__rollup__: _kwargs = context

      # resolve dispatch function
      dispatch = (func if not (
        isinstance(func, (classmethod, staticmethod))) else func.__func__)

      # notify function of hookname, if requested
      if self.__notify__: _args.insert(0, hookname)

      # dispatch
      return dispatch(*tuple(list(args) + _args), **_kwargs)
    return with_context



# -*- coding: utf-8 -*-

"""

  RPC exception tests
  ~~~~~~~~~~~~~~~~~~~

  tests canteen's builtin RPC systems.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
      A copy of this license is included as ``LICENSE.md`` in
      the root of the project.

"""



# -*- coding: utf-8 -*-

"""

  redis adapter tests
  ~~~~~~~~~~~~~~~~~~~

  tests canteen's redis adapter.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# redis adapter & model API
from canteen import model
from canteen.model.adapter import redis as rapi

# abstract test bases
from canteen_tests.test_adapters import test_abstract


try:
  import fakeredis
except ImportError:  # pragma: no cover
  fakeredis = None


if fakeredis:


  class RedisSetupTeardown(object):

    """ Packages Redis setup and teardown methods. """

    def test_adapter_mode(self):

      """ Test the adapter's internal mode against what we expect it to be """

      if not self.__abstract__:
        assert self.subject().__testing__ is True
        assert self.subject().EngineConfig.mode == self.mode

    def test_put_entity(self):

      """ Test saving a basic entity to Redis with `RedisAdapter` """


      class SampleEntity(model.Model):

        """ quick sample entity """

        __adapter__ = rapi.RedisAdapter

        string = str, {'indexed': False}
        number = int, {'indexed': False}

      s = SampleEntity(key=model.Key(SampleEntity, 'sample'),
                       string='hi',
                       number=5)
      x = s.put(adapter=self.subject())

      assert s.string == 'hi'
      assert s.number == 5
      assert x.kind == 'SampleEntity'
      assert x.id == 'sample'

      return s, x, SampleEntity

    def test_delete_entity(self):

      """ Test deleting a basic entity from Redis with `RedisAdapter` """

      s, x, SampleEntity = self.test_put_entity()
      s.delete(adapter=self.subject())

      ss = SampleEntity.get(x, adapter=self.subject())
      assert not ss, "should have deleted entity but instead got '%s'" % ss


  class RedisAdapterTopLevelBlobTests(test_abstract.DirectedGraphAdapterTests,
                                      RedisSetupTeardown):

    """ Tests `model.adapter.redis.Redis` in ``toplevel_blob`` mode """

    __abstract__ = False
    subject = rapi.RedisAdapter
    mode = rapi.RedisMode.toplevel_blob

    @classmethod
    def setUpClass(cls):
      """ Set Redis into testing mode. """

      rapi._mock_redis = fakeredis.FakeStrictRedis()
      rapi._mock_redis.flushall()
      rapi.RedisAdapter.__testing__ = True
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.toplevel_blob

    @classmethod
    def tearDownClass(cls):
      """ Set Redis back into non-testing mode. """

      rapi._mock_redis = fakeredis.FakeStrictRedis()
      rapi._mock_redis.flushall()
      rapi.RedisAdapter.__testing__ = False
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.toplevel_blob


  class RedisAdapterHashKindBlobTests(test_abstract.DirectedGraphAdapterTests,
                                      RedisSetupTeardown):

    """ Tests `model.adapter.redis.Redis` in ``hashkind_blob`` mode """

    __abstract__ = False
    subject = rapi.RedisAdapter
    mode = rapi.RedisMode.hashkind_blob

    @classmethod
    def setUpClass(cls):
      """ Set Redis into testing mode. """

      rapi.RedisAdapter.__testing__ = True
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.hashkind_blob

    @classmethod
    def tearDownClass(cls):
      """ Set Redis back into non-testing mode. """

      rapi.RedisAdapter.__testing__ = False
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.toplevel_blob


  class RedisAdapterHashKeyBlobTests(test_abstract.DirectedGraphAdapterTests,
                                     RedisSetupTeardown):

    """ Tests `model.adapter.redis.Redis` in ``hashkey_blob`` mode """

    __abstract__ = False
    subject = rapi.RedisAdapter
    mode = rapi.RedisMode.hashkey_blob

    @classmethod
    def setUpClass(cls):
      """ Set Redis into testing mode. """

      rapi.RedisAdapter.__testing__ = True
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.hashkey_blob

    @classmethod
    def tearDownClass(cls):
      """ Set Redis back into non-testing mode. """

      rapi.RedisAdapter.__testing__ = False
      rapi.RedisAdapter.EngineConfig.mode = rapi.RedisMode.toplevel_blob


  # @TODO(sgammon): add hashkey_hash testing when that mode is supported


else:  # pragma: no cover
  print("Warning! Redis not found, skipping Redis testsuite.")



# -*- coding: utf-8 -*-

"""

  page base
  ~~~~~~~~~

  ``Page``s represnt the simplest way to respond to an HTTP request in Canteen.
  They are inherently bound to HTTP, in that you specify methods directly from
  the HTTP spec to be executed when the corresponding method is requested.

  Example:

    # -*- coding: utf-8 -*-
    from canteen import url, Page

    @url('/')
    class Home(Page):


  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# DI & util
from . import handler


class Page(handler.Handler):

  """ Extendable class exposed to developers to prepare a class that responds to
      particular HTTP requests. Great way to return static content or render
      templates, as ``Page``s come preconfigured for use with :py:mod:`Jinja2`
      and Canteen's builtin logic (session, static asset and caching tools). """

  __owner__ = "Page"



# -*- coding: utf-8 -*-

"""

  core meta tests
  ~~~~~~~~~~~~~~~

  tests canteen's core, which contains abstract/meta code for constructing
  and gluing together the rest of canteen.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# testing
from canteen import test

# core meta
from canteen.core import meta
from canteen.core import injection
from canteen.core.meta import Proxy

# canteen util
from canteen.util import decorators


class CoreMetaTest(test.FrameworkTest):

  """ Tests for :py:mod:`canteen.core.meta` """

  def test_module_proxy(self):

    """ Test that metaproxy exposes `Factory`, `Registry` and `Component` """

    assert hasattr(meta, 'Proxy')
    assert hasattr(meta, 'MetaFactory')

  def test_proxy_attributes(self):

    """ Test that `meta.Proxy` properly exposes each meta-metaclass """

    assert hasattr(Proxy, 'Factory')
    assert hasattr(Proxy, 'Registry')
    assert hasattr(Proxy, 'Component')


class MetaFactoryTest(test.FrameworkTest):

  """ Tests for :py:class:`MetaFactory` """

  def test_new_concrete_metafactory(self):

    """ Test that `MetaFactory` cannot be constructed directly """

    with self.assertRaises(NotImplementedError):
      meta.MetaFactory()

    with self.assertRaises(NotImplementedError):
      meta.MetaFactory('hi')

    with self.assertRaises(NotImplementedError):
      meta.MetaFactory('hi', (object,))

  def test_new_meta_metafactory(self):

    """ Test that `MetaFactory` is usable as a metaclass via `Proxy.Factory` """

    assert isinstance(meta.Proxy.Factory('TargetMeta', (object,), {}), type)

  def test_meta_mro(self):

    """ Test that `meta.Proxy.Factory` properly overwrites object MRO """

    klass = meta.Proxy.Factory('TargetMeta', (object,), {
        '__metaclass__': meta.Proxy.Factory})
    assert klass.__mro__ == (klass, object)


class ClassFactoryTest(test.FrameworkTest):

  """ Tests for `meta.Proxy.Factory` """

  def test_construct_factory(self):

    """ Test constructing a new `meta.Proxy.Factory` meta-implementor """

    # make fake factory sub
    class FactoryMetaclass(object):
      __metaclass__ = meta.Proxy.Factory

    assert hasattr(FactoryMetaclass, 'initialize')
    assert not hasattr(FactoryMetaclass, 'inject')
    assert not hasattr(FactoryMetaclass, 'register')

  def test_initialize_factory(self):

    """ Test overriding `meta.Proxy.Factory.initialize` in a metaimplementor """

    # make subfactory
    class FactorySubclass(meta.Proxy.Factory):
      def initialize(cls, name, bases, properties):
        klass = type.__new__(cls, name, bases, properties)
        klass.__initialized__ = True
        return klass

    # make implementor
    class CoolImplementor(object):
      __metaclass__ = FactorySubclass

    assert hasattr(FactorySubclass, 'initialize')
    assert hasattr(CoolImplementor, '__initialized__')
    assert CoolImplementor.__initialized__ is True


class ClassRegistryTest(test.FrameworkTest):

  """ Tests for `meta.Proxy.Registry` """

  def test_construct_registry(self):

    """ Test constructing a new `meta.Proxy.Registry` metaimplementor """

    # make fake registry sub
    class RegistryMetaclass(object):
      __metaclass__ = meta.Proxy.Registry

    assert hasattr(RegistryMetaclass, 'register')
    assert hasattr(RegistryMetaclass, 'initialize')
    assert not hasattr(RegistryMetaclass, 'inject')

  def test_registry_internals(self):

    """ Test internals of `meta.Proxy.Registry` """

    # make a registered class tree
    class RegistryTestRegistry(object):
      __owner__, __metaclass__ = "RegistryTestRegistry", meta.Proxy.Registry

    # make implementors
    class RegisteredOne(RegistryTestRegistry): pass
    class RegisteredTwo(RegistryTestRegistry): pass

    # test chain internals
    registry = [i for i in (
        RegistryTestRegistry.__metaclass__.__chain__["RegistryTestRegistry"])]
    assert "RegistryTestRegistry" in (
        RegistryTestRegistry.__metaclass__.__chain__)
    assert RegisteredOne in registry
    assert RegisteredTwo in registry

  def test_iterate_children(self):

    """ Test iterating over a registered class' children """

    # make a registered class tree
    class IterateChildrenRegistry(object):
      __owner__, __metaclass__ = "IterateChildrenRegistry", meta.Proxy.Registry

    # make implementors
    class IterationChildOne(IterateChildrenRegistry): pass
    class IterationChildTwo(IterateChildrenRegistry): pass
    class IterationChildThree(IterateChildrenRegistry): pass

    _results = []
    for child in IterateChildrenRegistry.iter_children():
      _results.append(child)

    assert len(_results) == 3
    for i in (IterationChildOne, IterationChildTwo, IterationChildThree):
      assert i in _results

  def test_list_children(self):

    """ Test retrieving a list of a registered class children """

    # make a registered class tree
    class ListChildrenRegistry(object):
      __owner__, __metaclass__ = "ListChildrenRegistry", meta.Proxy.Registry

    # make implementors
    class ListChildOne(ListChildrenRegistry): pass
    class ListChildTwo(ListChildrenRegistry): pass
    class ListChildThree(ListChildrenRegistry): pass

    _results = ListChildrenRegistry.children()
    assert len(_results) == 3
    for i in (ListChildOne, ListChildTwo, ListChildThree):
      assert i in _results


class ClassComponentTest(test.FrameworkTest):

  """ Tests for :py:class:`meta.Proxy.Component` """

  def test_construct_component(self):

    """ Test constructing a new `meta.Proxy.Component` meta-implementor """

    # make fake component sub
    class ComponentMetaclass(object):
      __metaclass__ = meta.Proxy.Component

    assert hasattr(ComponentMetaclass, 'inject')
    assert hasattr(ComponentMetaclass, 'register')
    assert hasattr(ComponentMetaclass, 'initialize')

    return ComponentMetaclass

  def test_collapse(self):

    """ Test manually collapsing a class with known bindings """

    meta.Proxy.Component.reset_cache()


    @decorators.bind('injectable')
    class InjectableClass(self.test_construct_component()):

      """  """

      @decorators.bind('test', wrap=staticmethod)
      def test(self):

         """  """

      @decorators.bind('toplevel', wrap=staticmethod, namespace=False)
      def toplevel(self):

         """  """

    assert InjectableClass


    class TestCompound(object):
      __metaclass__ = injection.Compound


    class InjectedClass(TestCompound):

      """  """

      def test_blab(self):

        """  """

        assert self.toplevel
        assert self.injectable
        assert self.injectable.test

      def test_invalid(self):

        """  """

        self.invalid_property_here


    # manually perform a class collapse
    collapsed = meta.Proxy.Component.collapse(InjectedClass)

    # test collapse
    assert isinstance(collapsed, dict)
    assert 'injectable' in collapsed
    assert 'toplevel' in collapsed
    assert 'injectable.test' in collapsed

    # test full injection
    i = InjectedClass()
    assert i.test_blab
    i.test_blab()

    # make sure invalid attributes still fail
    with self.assertRaises(AttributeError):
      i.blabble

    with self.assertRaises(AttributeError):
      i.test_invalid()

    # clean up
    meta.Proxy.Component.reset_cache()

    return InjectedClass

  def test_singleton_map(self):

    """ Test setting a `Component` meta-implementor into `singleton` mode """


    @decorators.singleton
    @decorators.bind('singleton')
    class SingletonTest(self.test_construct_component()):

      """  """

      def get_self(self):

        """  """

        return self


    class TestCompound(object):

      """ Test compound object. """

      __metaclass__ = injection.Compound


    class CompoundSingletonTest(TestCompound):

      """  """

      def test_singleton(self):

        """  """

        assert self.singleton
        return self.singleton.get_self()

    # access the singleton to create one
    i = CompoundSingletonTest()
    singleton_one = i.test_singleton()
    singleton_two = i.test_singleton()

    # must be exactly equal
    assert singleton_one is singleton_two

    # check singleton's presence in the map
    assert singleton_one in SingletonTest.__class__.singleton_map.values()
    assert singleton_two in SingletonTest.__class__.singleton_map.values()



# -*- coding: utf-8 -*-

"""

  handler base
  ~~~~~~~~~~~~

  Presents a reasonable base class for a ``Handler`` object, which handles
  responding to an arbitrary "request" for action. For example, ``Handler``
  is useful for responding to HTTP requests *or* noncyclical realtime-style
  requests, and acts as a base class for ``Page`` and ``ServiceHandler``.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import itertools

# canteen core, util, logic
from ..core import injection


# noinspection PyUnresolvedReferences
class Handler(object):

  """ Base class structure for a ``Handler`` of some request or desired action.
      Specifies basic machinery for tracking a ``request`` alongside some form
      of ``response``.

      Also keeps track of relevant ``environ`` (potentially from WSGI) and sets
      up a jump off point for DI-provided tools like logging, config, caching,
      template rendering, etc. """

  # @TODO(sgammon): HTTPify, convert to decorator
  config = property(lambda self: {})

  __agent__ = None  # current `agent` details
  __status__ = 200  # it's a glass-half-full kind of day, why not
  __routes__ = None  # route map adapter from werkzeug
  __context__ = None  # holds current runtime context, if any
  __logging__ = None  # internal logging slot
  __runtime__ = None  # reference up to the runtime
  __environ__ = None  # original WSGI environment
  __request__ = None  # lazy-loaded request object
  __headers__ = None  # buffer HTTP header access
  __response__ = None  # lazy-loaded response object
  __callback__ = None  # callback to send data (sync or async)
  __content_type__ = None  # response content type

  # set owner and injection side
  __owner__, __metaclass__ = "Handler", injection.Compound

  def __init__(self, environ=None,
                     start_response=None,
                     runtime=None,
                     request=None,
                     response=None, **context):

    """ Initialize a new ``Handler`` object with proper ``environ`` details and
        inform it of larger world around it.

        ``Handler`` objects (much like ``Runtime`` objects) are designed to be
        usable independently as a WSGI-style callable. Note that the first two
        position parameters of this ``__init__`` are the venerable ``environ``
        and ``start_response`` - dispatching this way is totally possible, but
        providing ``runtime``, ``request`` and ``response`` allow tighter
        integration with the underlying runtime.

        Current execution details (internal to Canteen) are passed as ``kwargs``
          and compounded as new context items are added.

        :param environ: WSGI environment, provided by active runtime. ``dict``
          in standard WSGI format.

        :param start_response: Callable to begin the response cycle. Usually a
          vanilla ``function``.

        :param runtime: Currently-active Canteen runtime. Always an instance of
          :py:class:`canteen.core.runtime.Runtime` or a subclass thereof.

        :param request: Object to use for ``self.request``. Usually an instance
          of :py:class:`werkzeug.wrappers.Request`.

        :param response: Object to use for ``self.response``. Usually an
          instance of :py:class:`werkzeug.wrappers.Response`. """

    # startup/assign internals
    self.__runtime__, self.__environ__, self.__callback__ = (
      runtime,  # reference to the active runtime
      environ,  # reference to WSGI environment
      start_response)  # reference to WSGI callback

    # setup HTTP/dispatch stuff
    self.__status__, self.__headers__, self.__content_type__ = (
      200,  # default response status
      {},  # default repsonse headers
      'text/html; charset=utf-8')  # default content type

    # request, response & context
    self.__request__, self.__response__, self.__context__ = (
        request, response, context)

  # expose internals, but write-protect
  routes = property(lambda self: self.__runtime__.routes)
  status = property(lambda self: self.__status__)
  headers = property(lambda self: self.__headers__)
  content_type = property(lambda self: self.__content_type__)

  # shortcuts & utilities
  url_for = link = lambda self, end, **args: self.routes.build(end, args)

  # WSGI internals
  app = runtime = property(lambda self: self.__runtime__)
  environment = environ = property(lambda self: self.__environ__)
  start_response = callback = property(lambda self: self.__callback__)

  # Context
  session = property(lambda self: (  # session is tuple of (session, engine)
    self.request.session[0] if self.request.session else None))

  # Agent
  agent = property(lambda self: (
    self.__agent__ if self.__agent__ else (
      setattr(self, '__agent__', self.http.agent.scan(self.request)) or (
        self.__agent__))))

  # Request & Response
  request = property(lambda self: (
    self.__request__ if self.__request__ else (
      setattr(self, '__request__', self.http.new_request(self.__environ__)) or (
        self.__request__))))

  response = property(lambda self: (
    self.__response__ if self.__response__ else (
      setattr(self, '__response__', self.http.new_response()) or (
        self.__response__))))

  @property
  def template_context(self):

    """ Generate template context to be used in rendering source templates. The
        ``template_context`` accessor is expected to return a ``dict`` of
        ``name=>value`` pairs to present to the template API.

        :returns: ``dict`` of template context. """

    # for javascript context
    from canteen.rpc import ServiceHandler

    return {

      # Default Context
      'handler': self,
      'config': getattr(self, 'config', {}),
      'runtime': self.runtime,

      # HTTP Context
      'http': {
        'agent': getattr(self, 'agent', None),
        'request': self.request,
        'response': self.response
      },

      # WSGI internals
      'wsgi': {
        'environ': self.environ,
        'callback': self.callback,
        'start_response': self.start_response
      },

      # Cache API
      'cache': {
        'get': self.cache.get,
        'get_multi': self.cache.get_multi,
        'set': self.cache.set,
        'set_multi': self.cache.set_multi,
        'delete': self.cache.delete,
        'delete_multi': self.cache.delete_multi,
        'clear': self.cache.clear,
        'flush': self.cache.flush
      },

      # Assets API
      'asset': {
        'image': self.assets.image_url,
        'style': self.assets.style_url,
        'script': self.assets.script_url
      },

      # Service API
      'services': {
        'list': ServiceHandler.services,
        'describe': ServiceHandler.describe
      },

      # Output API
      'output': {
        'render': self.template.render,
        'environment': self.template.environment
      },

      # Routing
      'link': self.url_for,

      'route': {
        'build': self.url_for,
        'resolve': self.http.resolve_route
      }

    }

  def respond(self, content=None, direct=False):

    """ Respond to this ``Handler``'s request with raw ``str`` or ``unicode``
        content. UTF-8 encoding happens if necessary.

        :param content: Content to respond to. Must be ``str``, ``unicode``, or
          a similar string buffer object.

        :param direct: Flag indicating that ``self`` should be returned, rather
          than ``self.response``. Bool, defaults to ``False`` as this
          technically breaks WSGI.

        :returns: Generated (filled-in) ``self.response`` object. """

    # today is a good day
    if not self.status: self.__status__ = 200
    if content: self.response.response = content

    # set status code and return
    return setattr(self.response,
                  ('status_code' if isinstance(self.status, int) else 'status'),
                    self.status) or (
                  (i.encode('utf-8').strip() for i in self.response.response),
                        self.response) if not direct else self

  def render(self, template,
                   headers=None,
                   content_type='text/html; charset=utf-8',
                   context=None,
                   _direct=False, **kwargs):

    """ Render a source ``template`` for the purpose of responding to this
        ``Handler``'s request, given ``context`` and proper ``headers`` for
        return.

        ``kwargs`` are taken as extra template context and overlayed onto
        ``context`` before render.

        :param template: Path to template file to serve. ``str`` or ``unicode``
          file path.

        :param headers: Extra headers to send with response. ``dict`` or iter of
          ``(name, value)`` tuples.

        :param content_type: Value to send for ``Content-Type`` header. ``str``,
          defaults to ``text/html; charset=utf-8``.

        :param context: Extra template context to include during render.
          ``dict`` of items, with keys as names that values are bound to in the
          resulting template context.

        :param _direct: Flag indicating that ``self`` should be returned, rather
          than ``self.response``. Bool, defaults to ``False`` as this
          technically breaks WSGI.

        :param kwargs: Additional items to add to the template context.
          Overrides all other sources of context.

        :returns: Rendered template content, added to ``self.response``. """

    from canteen.util import config

    # set mime type
    if content_type: self.response.mimetype = content_type

    # collapse and merge HTTP headers (base headers first)
    self.response.headers.extend(itertools.chain(
      iter(self.http.base_headers),
      self.config.get('http', {}).get('headers', {}).iteritems(),
      self.headers.iteritems(),
      (headers or {}).iteritems()))

    # merge template context
    _merged_context = dict(itertools.chain(*(i.iteritems() for i in (
      self.template.base_context,
      self.template_context,
      context or {},
      kwargs))))

    # render template and set as response data
    self.response.response, self.response.direct_passthrough = (
      self.template.render(
        self,
        getattr(self.runtime, 'config', None) or config.Config(),
        template,
        _merged_context)), True

    return self.respond(direct=_direct)

  def dispatch(self, **url_args):

    """ Dispatch a WSGI request through this ``Handler``. Expected to be an
        HTTP-style (cyclical) dispatch flow.

        :param url_args: Arguments provided from the URI that should be passed
          along to any resulting handler calls.

        :returns: After filling the local response object (at ``self.response``)
          returns it for inspection or reply. """

    self.__response__ = (
      getattr(self, self.request.method)(**url_args)) or self.__response__
    return self.__response__

  def __call__(self, url_args, direct=False):

    """ Kick off the local response dispatch process, and run any necessary
        pre/post hooks (named ``prepare`` and ``destroy``, respectively).

        :param url_args: Arguments parsed from URL according to matched route.
          ``dict`` of ``{param: value}`` pairs.

        :param direct: Flag to indicate 'direct' mode, whereby a handler is
          returned instead of a response. Bool, defaults to ``False``, as this
          technically breaks WSGI.

        :returns: ``self.response`` if ``direct`` mode is not active, otherwise
          ``self`` for chainability. """

    # run prepare hook, if specified
    if hasattr(self, 'prepare'): self.prepare(url_args, direct=direct)

    self.dispatch(**url_args)  # dispatch local handler, fills `__response__`

    # run destroy hook, if specified
    if hasattr(self, 'destroy'): self.destroy(self.__response__)
    return self.__response__ if not direct else self


# noinspection PyUnresolvedReferences
class RealtimeHandler(Handler):

  """ Provides structure for an acyclically-dispatched web handler, meant for
      use in scenarios like WebSockets. Instead of handling things with
      methods like ``GET`` or ``POST``, a ``RealtimeHandler`` can specify
      hooks for two events - ``on_connect`` and ``on_message``.

      The first, ``on_connect``, is dispatched when a realtime connection has
      just been successfully negotiated. It is executed once the application
      is ready to return an ``HTTP/1.1 Upgrade`` response, so that the
      developer has a chance to specify subprotocols/extensions/etc.

      The second hook, ``on_message``, is dispatched each time an established
      connection receives a message from the client. It takes two parameters -
      the ``message`` itself and whether it is ``binary`` or not. """

  __socket__ = None  # space for currently-active realtime socket

  def dispatch(self, **url_args):  # pragma: no cover

    """ Adapt regular handler dispatch to support an acyclic/realtime-style
        dispatch scheme. Accepts same arguments as ``super`` definition, but
        dispatches *realtime*-style messages like ``on_connect`` and
        ``on_message``, so long as the request looks like a WebSocket upgrade.

        :param url_args: Arguments provided from the URI that should be passed
          along to any resulting handler calls.

        :returns: After filling the local response object (at ``self.response``)
          returns it for inspection or reply. """

    # fallback to standard dispatch
    if self.realtime.hint not in self.environ:
      return super(RealtimeHandler, self).dispatch(**url_args)

    try:
      # websocket upgrade and session
      self.__socket__ = self.realtime.on_connect(self)
      self.realtime.on_message(self, self.__socket__)

    except NotImplementedError:
      return self.error(400)  # raised when a non-websocket handler is hit

  @staticmethod
  def terminate(graceful=True):  # pragma: no cover

    """ Terminate the currently-active ``RealtimeSocket`` communication
        channel.

        :param graceful: ``bool`` parameter, whether to end the connection
          gracefully or not.

        :returns: ``TERMINATE`` sentinel, to be yielded so the connection can be
          terminated. """

    from canteen.logic import realtime

    if graceful: return realtime.TERMINATE
    raise realtime.TerminateSocket(graceful=False)

  @staticmethod
  def on_connect():  # pragma: no cover

    """ Hook function that is dispatched upon successful handshake for a
        realtime-style connection between a client and this server. Local
        handler should be prepared by this point with all information necessary
        to satisfy messages.

        Implementors are expected to provide a method that makes use of object-
        level context (i.e. not a static or classmethod).

        :returns: ``NotImplemented`` by default, which simply indicates that
          the implementor elects not to run code ``on_connect``. """

    return NotImplemented

  def on_message(self, message, binary):  # pragma: no cover

    """ Hook that is dispatched per message sent from a live client. Called
        subsequent to a connection being properly established from a previous
        call to ``on_connect``.

        :param message: WebSocket message passed from the client.

        :param binary: ``bool`` flag - ``True`` if ``message`` is binary,
          ``False`` otherwise.

        :raises NotImplementedError: By default, since not many people use
          WebSockets and there's no such thing as a ``400`` without HTTP. :)

        :returns: Not expected to return anything. If a return is used, any
          value or iterable of values will be collapsed and sent to the client.
          Optionally, the developer may implement ``on_message`` as a coroutine-
          style Python generator, in which case new messages will be ``sent``
          in from the client and messages to the client can be yielded upwards
          to be sent. """

    raise NotImplementedError('Handler "%s" fails to implement hook'
                              ' `on_message` so it does not support'
                              ' realtime-style communications.' % repr(self))

  # noinspection PyUnusedLocal
  @staticmethod
  def on_close(graceful):  # pragma: no cover

    """ Hook function that is dispatched upon closure of an existing realtime
        communications session.

        :param graceful: ``bool`` parameter indicating whether the connection
          was closed gracefully (i.e. electively) or because of some error
          condition.

        :returns: ``NotImplemented`` by default, which simply indicates that
          the implementor elects not to run code ``on_connect``. """

    return NotImplemented


__all__ = ('Handler',)



# -*- coding: utf-8 -*-

"""

  assets logic
  ~~~~~~~~~~~~

  exposes logic for managing and making-use-of static assets.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import os
import random
import hashlib
import mimetypes

# core, base & util
from ..base import logic
from ..core import hooks
from ..util import config
from ..util import decorators


## Globals
_default_asset_path = os.path.join(os.getcwd(), 'assets')


@decorators.bind('assets')
class Assets(logic.Logic):

  """ Provides logic related to the use and management of static assets, such as
      images, CSS, JavaScript, fonts, etc.

      Canteen enables two styles of asset management: *registered* and
      *unregistered*. When using registered assets, developers add their static
      files to config (for everything but images) and use a dereferenced naming
      scheme to reference them from templates or handlers. """

  __config__ = None  # asset configuration, if any
  __handles__ = {}  # cached file handles for local responders
  __prefixes__ = {}  # static asset type prefixes
  __static_types__ = frozenset(('style', 'script', 'font', 'image', 'video'))

  ### === Internals === ###
  debug = property(lambda self: self.config.get('debug', True))
  assets = property(lambda self: config.Config().assets.get('assets', {}))

  config = property(lambda self: (
    config.Config().assets.get('config', {'debug': True})))

  path = property(lambda self: (
    config.Config().app.get('paths', {}).get('assets', _default_asset_path)))

  ### === Detection & Bindings === ###
  @hooks.HookResponder('initialize')
  def bind_urls(self):

    """ Bind static asset URLs, if Canteen is instructed to handle requests for
        static assets. Constructs handlers according to URLs and paths from
        application configuration. """

    from canteen import url, handler

    ## asset handler
    def make_responder(asset_type, path_prefix=None):

      """ Internal utility function to make an ``AssetResponder`` handler which
          can handler assets of type ``asset_type`` at prefix ``path_prefix``.

          :param asset_type: Type of asset we're binding this handler for.
          :param path_prefix: URL path prefix that we should respond to.

          :returns: :py:class:`AssetResponder` instance, which is a Canteen
            :py:class:`base.Handler`, preconfigured to handle asset URLs. """


      class AssetResponder(handler.Handler):

        """ Internal :py:class:`canteen.base.Handler` implementation for binding
            to and fulfilling static asset URLs, such as those for CSS, images,
            JavaScript files and SVGs. """

        content_types = {
          'css': 'text/css',
          'js': 'application/javascript',
          'svg': 'image/svg+xml',
          'woff': 'font/woff',
          'png': 'image/png',
          'gif': 'image/gif',
          'jpeg': 'image/jpeg',
          'jpg': 'image/jpeg',
          'webp': 'image/webp',
          'webm': 'video/webm',
          'avi': 'video/avi',
          'mpeg': 'video/mpeg',
          'mp4': 'video/mp4',
          'flv': 'video/x-flv',
          'appcache': 'text/cache-manifest'}

        def GET(self, asset):

          """ Fulfill HTTP GET requests for a particular kind of ``asset_type``
              and ``path_prefix``, which are provided by the outer closure that
              constructs this handler.

              :param asset: Asset to serve. This is a relative path that should
                be translated into a local file and served.

              :returns: Response containing the headers and content to be served
                for the resource specified at ``asset``. """

          fullpath = (
            os.path.join(path_prefix, asset) if path_prefix else (
              os.path.join(self.assets.path, asset_type, asset)))
          if fullpath in self.assets.__handles__:

            # extract cached handle/modtime/content
            modtime, handle, contents, fingerprint = (
              self.assets.__handles__[fullpath])

            if os.path.getmtime(fullpath) > modtime:
              modtime, handle, contents, fingerprint = (
                self.open_and_serve(fullpath))  # need to refresh cache

          else:
            modtime, handle, contents, fingerprint = (
              self.open_and_serve(fullpath))  # need to prime cache

          # try to serve a 304, if possible
          if 'If-None-Match' in self.request.headers:

            # fingerprint matches, serve a 304
            if self.request.headers['If-None-Match'] == fingerprint:
              etag_header = ('ETag', self.request.headers['If-None-Match'])
              return self.http.new_response(
                        status='304 Not Modified',
                        headers=[etag_header])

          # resolve content type by file extension, if possible
          content_type = self.content_types.get(fullpath.split('.')[-1])
          if not content_type:

            # try to guess with `mimetypes`
            content_type, encoding = mimetypes.guess_type(fullpath)
            if not content_type: content_type = 'application/octet-stream'

          # can return content directly
          return self.http.new_response(contents,
                                        headers=[('ETag', fingerprint)],
                                        content_type=content_type)

        # noinspection PyBroadException
        def open_and_serve(self, filepath):

          """ Utility function to open a static asset file and serve its
              contents. Takes a filepath and properly handles MIME/content-type
              negotiation before responding with the asset.

              :param filepath: Path to the static asset to be served.

              :returns: Structure (``tuple``) containing:
                - a timestamp for when the asset was last modified, provided by
                  ``getmtime`` from the OS
                - original handle to the file, which should be closed after
                  exiting this function
                - the contents of the file, as a string
                - an MD5 hash of the contents of the file, as a hex digest """

          if os.path.exists(filepath):
            try:
              with open(filepath, 'rb') as fhandle:

                # assign to cache location by file path
                contents = fhandle.read()
                self.assets.__handles__[filepath] = (
                  os.path.getmtime(filepath),
                  fhandle,
                  contents,
                  hashlib.md5(contents).hexdigest())
                return self.assets.__handles__[filepath]

            except IOError:
              if __debug__: raise
              self.error(404)

            except Exception:
              if __debug__: raise
              self.error(500)

          else:
            return self.error(404)

      return AssetResponder

    # set default asset prefixes
    asset_prefixes = self.__prefixes__ = {
      'style': 'assets/style',
      'image': 'assets/img',
      'script': 'assets/script',
      'font': 'assets/font',
      'video': 'assets/video',
      'other': 'assets/ext'
    } if 'asset_prefix' not in self.config else self.config['asset_prefix']

    for category, prefix in asset_prefixes.iteritems():
      url("%s-assets" % category, "/%s/<path:asset>" % prefix)((
        make_responder(asset_type=category)))

    if 'extra_assets' in self.config:
      for name, ext_cfg in self.config['extra_assets'].iteritems():
        prefix, path = ext_cfg
        url("%s-extra-assets" % name, "%s/<path:asset>" % prefix)((
          make_responder(asset_type=name, path_prefix=path)))

  ### === Resolvers === ###
  def find_filepath(self, asset_type):

    """  """

    if isinstance(self.path, dict):
      if asset_type in self.path:
        return (self.path['asset_type'], asset_type)
    return (self.path, asset_type)

  def find_path(self, asset_type):

    """  """

    if isinstance(self.__prefixes__, dict):
      # allow type-specific asset prefixes
      if asset_type in self.__prefixes__:
        return self.__prefixes__[asset_type]
      raise ValueError("Cannot calculate asset prefix"
                       " for unspecified asset type '%s'." % asset_type)
    return ("assets/%s" % asset_type, asset_type)

  ### === URL Builders === ###
  def asset_url(self, type, fragments, arguments):

    """  """

    assert (fragments or arguments)  # must pass at least fragments or arguments

    if type not in self.__static_types__:
      raise ValueError("Cannot generate asset URL"
                       " for unknown asset type '%s'." % type)

    if fragments:
      url_blocks = {

        # one fragment will be a relative URL, like "sample/app.css?v1"
        1: lambda relative: (self.find_path(type), relative),

        # two fragments is a package + name
        2: lambda package, name: self.find_name(type, package, name),

        # three fragments is a package + name + version
        3: lambda name, package, version: self.find_name(*(
          type, package, name), version=version)

      }.get(len(fragments))(*fragments)

    if arguments:
      # @TODO(sgammon): make this named-parameter friendly
      raise RuntimeError('Keyword-based asset URLs are not yet supported.')

    # build relative URL unless specified otherwise
    if arguments.get('absolute'):
      pass

    # should we use a CDN prefix?
    prefix = ''
    if self.config.get('serving_mode', 'local') == 'cdn':
      prefix += random.choice(self.config.get('cdn_prefix')) if (
        isinstance(self.config.get('cdn_prefix', None), (list, tuple))) else (
          self.config.get('cdn_prefix'))

    return prefix + '/'.join([''] + ['/'.join(map(lambda x: '/'.join(x) if (
      isinstance(x, tuple)) else x, url_blocks))])

  @decorators.bind()  # CSS
  def style_url(self, *fragments, **arguments):

    """ Exports logic to calculate a URL for a *regsitered* or *unregistered*
        static CSS asset from a set of spec ``fragments`` and ``arguments``.

      :param fragments: Positional specification fragments that should be used
        to resolve the asset to be linked-to.

      :param arguments: Keyword arguments to be used for filling in path items
        when building the CSS asset URL.

      :return: Generated CSS asset URL. """

    return (
    self.asset_url('style', fragments, arguments))

  @decorators.bind()  # JS
  def script_url(self, *fragments, **arguments): return (
    self.asset_url('script', fragments, arguments))

  @decorators.bind()  # Fonts
  def font_url(self, *fragments, **arguments): return (
    self.asset_url('font', fragments, arguments))

  @decorators.bind()  # Images
  def image_url(self, *fragments, **arguments): return (
    self.asset_url('image', fragments, arguments))

  @decorators.bind()  # Video
  def video_url(self, *fragments, **arguments): return (
    self.asset_url('video', fragments, arguments))

  @decorators.bind()  # Other
  def static_url(self, *fragments, **arguments): return (
    self.asset_url('static', fragments, arguments))



# -*- coding: utf-8 -*-

"""

  framework setup
  ~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""


if __debug__:

  # setuptools
  import sys
  import logging
  import traceback
  import setuptools as tools


  ## Constants / Globals

  dependencies = []

  # version of setuptools needed for a pleasant life
  SETUPTOOLS_VERSION = '4.0.1'

  # custom version to forcefail pypi
  HAMLISH_VERSION, PROTOBUF_VERSION = '0.3.4-canteen', '2.5.2-canteen'

  # extract float of major version in X.X[.X]
  VERSION = lambda version: float('.'.join(version.split('.')[0:1]))

  # check version of setuptools
  CHECK_SETUPTOOLS = lambda: CURRENT_SETUPTOOLS_VERSION <= SETUPTOOLS_VERSION

  # extract current seutptools version
  CURRENT_SETUPTOOLS_VERSION = VERSION(getattr(tools, '__version__', '0.0.0'))

  # git endpoint for customized protobuf (formats with version)
  PROTOBUF_GIT = "git+git://github.com/sgammon/protobuf.git" \
                 "#egg=protobuf-%s"

  # git endpoint for customized hamlish (formats with version)
  HAMLISH_GIT = "git+git://github.com/sgammon/hamlish-jinja.git" \
                "#egg=hamlish_jinja-%s"

  ## Logging
  log = logging.getLogger('canteen.setup')
  log_handler = logging.StreamHandler(sys.stdout)
  log.addHandler(log_handler), log.setLevel((
    logging.DEBUG if __debug__ else logging.WARNING))


  def prepare():  # pragma: no cover

    """ Prepare constants and tools for setting up Canteen.

        :returns: ``go``, a closured function that, when called, will begin
          framework setup. """

    try:
      from colorlog import ColoredFormatter
    except ImportError:  # pragma: no cover
      log.debug('No support found for `colorlog`. No colors for u.')
    else:  # pragma: no cover
      log_handler.setFormatter(ColoredFormatter(
        "%(log_color)s[%(levelname)s]%(reset)s %(message)s",
        datefmt=None, reset=True, log_colors={
          'DEBUG':    'cyan',
          'INFO':     'green',
          'WARNING':  'yellow',
          'ERROR':    'red',
          'CRITICAL': 'red'}))


    ## Environment checks
    if not CHECK_SETUPTOOLS():
      log.warning('Setuptoo¡s out of date with version "%s", where'
                  'at least version "%s" is required.'
                  ' Attempting upgrade...' % (SETUPTOOLS_VERSION))
      dependencies.append('setuptools<=%s' % SETUPTOOLS_VERSION)

      try:
        # force update/download
        from ez_setup import use_setuptools
        use_setuptools()

        # reload module and check version
        reload(tools)
        CURRENT_SETUPTOOLS_VERSION = tools.__version__
      except Exception as e:
        log.error('Encountered exception using `ez_setup`...')
        if __debug__:
          traceback.print_exc()

      # fail hard if no suitable options
      if not CHECK_SETUPTOOLS():
        log.error('Failed to find a suitable version of setuptools.'
                  ' Building without support for HAML or RPC.')
        sys.exit(1)

    try:
      import protorpc
    except ImportError:
      log.info('Protobuf not found.'
               ' Adding custom version "%s"...' % PROTOBUF_VERSION)
      dependencies.append('protobuf==%s' % PROTOBUF_VERSION)

    try:
      import hamlish_jinja
    except ImportError:
      log.info('HamlishJinja requested but not found.'
               ' Adding custom version "%s"...' % HAMLISH_VERSION)
      dependencies.append('hamlish_jinja==%s' % HAMLISH_VERSION)

    return lambda: tools.setup(

      # == info == #
      name="canteen",
      version="0.3-alpha",
      description="Minimally complicated, maximally blasphemous"
                  " approach to Python development",

      # == authorship == #
      author="Sam Gammon",
      author_email="sam@momentum.io",

      # == flags == #
      zip_safe=True,
      url="https://github.com/sgammon/canteen",

      # == package tree == #
      packages=["canteen",
                "canteen.base",
                "canteen.core",
                "canteen.logic",
                "canteen.logic.http",
                "canteen.model",
                "canteen.model.adapter",
                "canteen.rpc",
                "canteen.rpc.protocol",
                "canteen.runtime",
                "canteen.util"] +

               ["canteen_tests",
                "canteen_tests.test_adapters",
                "canteen_tests.test_base",
                "canteen_tests.test_core",
                "canteen_tests.test_http",
                "canteen_tests.test_model",
                "canteen_tests.test_rpc",
                "canteen_tests.test_util"] if __debug__ else [],

      # == dependencies == #
      install_requires=([
                          "jinja2",
                          "werkzeug",
                          "protorpc"] + dependencies),

      # == test dependencies == #
      tests_require=("nose", "coverage", "fakeredis"),

      # == dependency links == #
      dependency_links=(
        PROTOBUF_GIT % PROTOBUF_VERSION,
        HAMLISH_GIT % HAMLISH_VERSION))



# -*- coding: utf-8 -*-

"""

  base
  ~~~~

  This package holds classes that extend framework internals into usable base
  classes. Developers extend classes in this package to create an application.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# submodules
from .page import *
from .logic import *
from .handler import *
from .protocol import *


__all__ = ('page', 'Page',
           'logic', 'Logic',
           'handler', 'Handler',
           'protocol', 'Protocol')



# -*- coding: utf-8 -*-

"""

  stdlib runtime
  ~~~~~~~~~~~~~~

  runs :py:mod:`canteen`-based apps on python's stdlib library,
  :py:mod:`wsgiref`.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# core
from ..core import runtime


with runtime.Library('wsgiref') as (library, wsgiref):

  # stdlib
  simple_server = library.load('simple_server')


  class StandardWSGI(runtime.Runtime):

    """  """

    __default__ = True

    def bind(self, interface, port):  # pragma: no cover

      """  """

      return simple_server.make_server(interface, port, self.dispatch)



# -*- coding: utf-8 -*-

"""

  model query tests
  ~~~~~~~~~~~~~~~~~

  tests querying in canteen's model layer.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# canteen
from canteen import model
from canteen.model import query
from canteen.test import FrameworkTest

# inmemory adapter
from canteen_tests.test_adapters import test_inmemory as inmemory


## QueryTests
class QueryTests(FrameworkTest):

  """ Tests for `model.query.Query` and `model.query.QueryOptions`. """

  def test_construct(self):

    """ Test basic construction of a `model.Query` """

    assert hasattr(query, 'Query')
    assert query.Query
    assert query.Query()

  def test_options(self):

    """ Test basic construction of `query.QueryOptions` """

    assert hasattr(query, 'QueryOptions')
    assert query.QueryOptions()

  def test_options_invalid(self):

    """ Test setting an invalid setting on `QueryOptions` """

    options = query.QueryOptions()
    with self.assertRaises(AttributeError):
      options._set_option('make_me_dinner', True)

  def test_options_limit(self):

    """ Test the use of a query limit in `QueryOptions`  """

    assert query.QueryOptions(limit=50).limit == 50

  def test_options_offset(self):

    """ Test the use of a query offset in `QueryOptions` """

    assert query.QueryOptions(offset=50).offset == 50

  def test_options_ancestor(self):

    """ Test the use of a query ancestor in `QueryOptions` """

    assert query.QueryOptions(ancestor=model.Key('SampleModel', 'hi')).ancestor

  def test_options_overlay(self):

    """ Test overlaying one `QueryOptions` on another """

    left = query.QueryOptions(limit=50)
    right = query.QueryOptions(ancestor=model.Key('SampleModel', 'hi'))

    combined = left.overlay(right)
    assert combined.limit == 50
    assert combined.ancestor == model.Key('SampleModel', 'hi')

  def test_options_overlay_override(self):

    """ Test overlaying property overrides with `QueryOptions` """

    left = query.QueryOptions(limit=50)
    right = query.QueryOptions(limit=25,
                               ancestor=model.Key('SampleModel', 'hi'))

    overridden = left.overlay(right, override=True)
    assert overridden.limit == 25
    assert overridden.ancestor == model.Key('SampleModel', 'hi')

  def test_options_overlay_no_override(self):

    """ Test safely overlaying properties with `QueryOptions` """

    left = query.QueryOptions(limit=50)
    right = query.QueryOptions(limit=25,
                               ancestor=model.Key('SampleModel', 'hi'))

    overridden = left.overlay(right, override=False)
    assert overridden.limit == 50
    assert overridden.ancestor == model.Key('SampleModel', 'hi')

  def test_optioned_query(self):

    """ Test using `QueryOptions` with a `model.Query` """

    options = query.QueryOptions(limit=50,
                                 ancestor=model.Key('SampleModel', 'hi'))
    q = query.Query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key('SampleModel', 'hi')

  def test_interface_primitive_filter_objects(self):

    """ Test using primitive `query.Filter` objects in a `model.Query` """

    class SomeModel(model.Model):

        """ Some sample model """

        string = basestring
        integer = int

    filter_one = query.Filter(SomeModel.string, 'hi', operator=query.EQUALS)
    filter_two = query.Filter(SomeModel.integer, 5, operator=query.GREATER_THAN)

    q = SomeModel.query(filter_one, filter_two)

    assert len(q.filters) == 2
    assert filter_one in q.filters
    assert filter_two in q.filters

  def test_interface_primitive_sort_objects(self):

    """ Test using primitive `query.Sort` objects in a `model.Query` """

    class SomeModel(model.Model):

        """ Some sample model """

        string = basestring
        integer = int

    sort_one = query.Sort(SomeModel.string, operator=query.ASCENDING)
    sort_two = query.Sort(SomeModel.integer, operator=query.DSC)

    q = SomeModel.query(sort_one, sort_two)

    assert len(q.sorts) == 2
    assert sort_one in q.sorts
    assert sort_two in q.sorts

  def test_interface_primitive_both_objects(self):

    """ Test primitive `query.Sort` inline with `query.Filter` objects """

    class SomeModel(model.Model):

        """ Some sample model """

        string = basestring
        integer = int

    filter_one = query.Filter(SomeModel.string, 'hi', operator=query.EQUALS)
    filter_two = query.Filter(SomeModel.integer, 5, operator=query.GREATER_THAN)
    sort_one = query.Sort(SomeModel.string, operator=query.ASCENDING)
    sort_two = query.Sort(SomeModel.integer, operator=query.DSC)

    q = SomeModel.query(filter_one, filter_two, sort_one, sort_two)

    assert len(q.filters) == 2
    assert filter_one in q.filters
    assert filter_two in q.filters

    assert len(q.sorts) == 2
    assert sort_one in q.sorts
    assert sort_two in q.sorts

  def test_interface_sort_default(self):

    """ Test specifying a (default-direction) `Sort` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.sort(-inmemory.InMemoryModel.string)

    assert len(q.sorts) == 1
    assert q.sorts[0].operator == query.DESCENDING

  def test_interface_sort_ascending(self):

    """ Test specifying an ascending `Sort` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.sort(+inmemory.InMemoryModel.string)

    assert len(q.sorts) == 1
    assert q.sorts[0].operator == query.ASCENDING

  def test_interface_sort_descending(self):

    """ Test specifying a descending `Sort` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.sort(-inmemory.InMemoryModel.string)

    assert len(q.sorts) == 1
    assert q.sorts[0].operator == query.DESCENDING

  def test_interface_equality_filter(self):

    """ Test specifying an equality `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.string == 'sup')

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.EQUALS

  def test_interface_inequality_filter(self):

    """ Test specifying an inequality `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.string != 'sup')

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.NOT_EQUALS

  def test_interface_greater_than_filter(self):

    """ Test specifying a greater-than `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.integer > 5)

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.GREATER_THAN

  def test_interface_less_than_filter(self):

    """ Test specifying a less-than `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.integer < 5)

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN

  def test_interface_greater_than_equal_to_filter(self):

    """ Test specifying a greater-than-equal-to `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.integer >= 5)

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.GREATER_THAN_EQUAL_TO

  def test_interface_less_than_equal_to_filter(self):

    """ Test specifying a less-than-equal-to `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.number <= 5)

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO

  def test_manual_filter_match(self):

    """ Test manually matching against a `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.number <= 5)
    matching_model = inmemory.InMemoryModel(number=1, string='womp')

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model)

    q = inmemory.InMemoryModel.query(options=options)
    q.filter((inmemory.InMemoryModel.number <= 5).AND((
        inmemory.InMemoryModel.string == 'womp')))

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model)

    q = inmemory.InMemoryModel.query(options=options)
    q.filter((inmemory.InMemoryModel.number <= 5).OR((
        inmemory.InMemoryModel.string == 'womp')))

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model)

  def test_manual_filter_match_value(self):

    """ Test manually matching a raw value against a `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.number <= 5)
    matching_model = inmemory.InMemoryModel(number=1, string='womp')

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model.number)

  def test_manual_filter_match_raw_entity(self):

    """ Test manually matching a raw entity against a `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.number <= 5)
    matching_model = inmemory.InMemoryModel(number=1, string='womp')

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model.to_dict())

  def test_sort_string_repr(self):

    """ Test the string representation for a `Sort` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.sort(-inmemory.InMemoryModel.string)

    assert len(q.sorts) == 1
    assert q.sorts[0].operator == query.DESCENDING

    result = q.sorts[0].__repr__()
    assert 'Sort' in result
    assert 'DESCENDING' in result

  def test_filter_string_repr(self):

    """ Test the string representation for a `Filter` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    q = inmemory.InMemoryModel.query(options=options)

    assert q.options.limit == 50
    assert q.options.ancestor == model.Key(inmemory.InMemoryModel, 'hi')

    q.filter(inmemory.InMemoryModel.integer <= 5)
    matching_model = inmemory.InMemoryModel(integer=1)

    assert len(q.filters) == 1
    assert q.filters[0].operator == query.LESS_THAN_EQUAL_TO
    assert q.filters[0].match(matching_model)

    result = q.filters[0].__repr__()
    assert 'Filter' in result
    assert 'integer' in result

  def test_options_string_repr(self):

    """ Test the string representation for a `QueryOptions` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    result = options.__repr__()

    assert 'QueryOptions' in result
    assert 'limit' in result
    assert 'ancestor' in result

  def test_query_string_repr(self):

    """ Test the string representation for a `Query` """

    _ancestor = model.Key(inmemory.InMemoryModel, 'hi')
    options = query.QueryOptions(limit=50, ancestor=_ancestor)
    _q = inmemory.InMemoryModel.query(options=options)
    _q.filter(inmemory.InMemoryModel.integer <= 5)
    _q.sort(+inmemory.InMemoryModel.string)

    result = _q.__repr__()

    assert 'limit' in result
    assert 'ancestor' in result
    assert 'Sort' in result
    assert 'Filter' in result
    assert 'ASCENDING' in result
    assert '<=' in result

  def test_chained_construction(self):

    """ Test constructing a chained query with `Query` """

    class SomeModel(model.Model):

        """ Some sample model """

        string = basestring
        integer = int

    filter_one = query.Filter(SomeModel.string, 'hi', operator=query.EQUALS)
    filter_two = query.Filter(SomeModel.integer, 5,
                              operator=query.GREATER_THAN,
                              AND=(filter_one,))

    q = SomeModel.query(filter_one, filter_two)

    assert len(q.filters) == 2
    assert filter_one in q.filters
    assert filter_two in q.filters

    q = query.Query(filters=filter_one)

  def test_repeated_contains(self):

    """ Test constructing a CONTAINS query w/EQUALS on a repeated property """

    class SomeModel(model.Model):

        """ Some sample model """

        string = basestring, {'repeated': True}
        integer = int

    filter_one = query.Filter(SomeModel.string, 'hi', operator=query.EQUALS)
    q = SomeModel.query(filter_one)

    assert len(q.filters) == 1
    assert filter_one in q.filters
    assert filter_one.operator is query.CONTAINS



# -*- coding: utf-8 -*-

"""

  canteen: tests
  ~~~~~~~~~~~~~~

  class structure and testsuite to put canteen functionality through
  unit/integration/functional-level testing.

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

if __debug__:

  # test tools
  from canteen import test

  # import testsuite
  from . import test__init__
  from . import test_core
  from . import test_dispatch
  from . import test_rpc
  from . import test_test
  from . import test_util
  from . import test_model
  from . import test_adapters


  class SanityTest(test.FrameworkTest):

    """ Run some basic sanity tests. """

    def test_math_sanity(self):

      """ Test that math still works (lol) """

      self.assertEqual(1 + 1, 2)
      self.assertEqual(2 - 1, 1)

      assert (10 / 5) == 2
      assert (10 % 6) == 4

    def test_assert_sanity(self):

      """ Test `assert` behavior """

      try:
        assert 1 == 2
      except AssertionError:
        pass
      else:  # pragma: no cover
        raise RuntimeError('Assertions are disabled. Something is wrong, '
                           ' as `__debug__` is truthy.')



# -*- coding: utf-8 -*-

"""

  protorpc model extensions
  ~~~~~~~~~~~~~~~~~~~~~~~~~

  :author: Sam Gammon <sg@samgammon.com>
  :copyright: (c) Sam Gammon, 2014
  :license: This software makes use of the MIT Open Source License.
            A copy of this license is included as ``LICENSE.md`` in
            the root of the project.

"""

# stdlib
import datetime

# adapter API
from .abstract import KeyMixin
from .abstract import ModelMixin

# canteen util
from canteen.util import struct as datastructures


## == protorpc support == ##
try:
  # force absolute import to prevent infinite recursion
  protorpc = __import__('protorpc', tuple(), tuple(), [], 0)

except ImportError as e:  # pragma: no cover
  # flag as unavailable
  _PROTORPC, _root_message_class = False, object

else:
  # extended imports (must be absolute)
  _p = __import__('protorpc', [], [], ['messages', 'message_types'], 0)
  pmessages = getattr(_p, 'messages')
  pmessage_types = getattr(_p, 'message_types')

  # constants
  _model_impl = {}
  _field_kwarg = 'field'
  _PROTORPC, _root_message_class = True, pmessages.Message

  # map fields to basetypes
  _field_basetype_map = {
    int: pmessages.IntegerField,
    long: pmessages.IntegerField,
    bool: pmessages.BooleanField,
    float: pmessages.FloatField,
    str: pmessages.StringField,
    unicode: pmessages.StringField,
    basestring: pmessages.StringField,
    datetime.time: pmessages.StringField,
    datetime.date: pmessages.StringField,
    datetime.datetime: pmessages.StringField}

  # build quick basetype lookup
  _builtin_basetypes = frozenset(_field_basetype_map.keys())

  # map fields to explicit names
  _field_explicit_map = {
    pmessages.EnumField.__name__: pmessages.EnumField,  # 'EnumField'
    pmessages.BytesField.__name__: pmessages.BytesField,  # 'BytesField'
    pmessages.FloatField.__name__: pmessages.FloatField,  # 'FloatField'
    pmessages.StringField.__name__: pmessages.StringField,  # 'StringField'
    pmessages.IntegerField.__name__: pmessages.IntegerField,  # 'IntegerField'
    pmessages.BooleanField.__name__: pmessages.BooleanField}  # 'BooleanField'

  # build quick builtin lookup
  _builtin_fields = frozenset(_field_explicit_map.keys())


  # recursive message builder
  def build_message(_model):

    """ Recursively builds a new `Message` class dynamically from a canteen
        :py:class:`model.Model`. Properties are converted to their
        :py:mod:`protorpc` equivalents and factoried into a full
        :py:class:`messages.Message` class.

        :param _model: Model class to convert to a
          :py:class:`protorpc.messages.Message` class.

        :raises TypeError: In the case of an unidentified or unknown
          property basetype.
        :raises ValueError: In the case of a missing implementation field
          or serialization error.

        :returns: Constructed (but not instantiated)
          :py:class:`protorpc.messages.Message` class. """

    # must nest import to avoid circular dependencies
    from canteen import rpc
    from canteen import model

    # provision field increment and message map
    _field_i, _model_message = 1, {'__module__': _model.__module__}

    # grab lookup and property dict
    lookup, property_map = _model.__lookup__, {}

    # add key submessage
    _model_message['key'] = pmessages.MessageField(rpc.Key, _field_i)

    # build fields from model properties
    for name in lookup:

      # init args and kwargs
      _pargs, _pkwargs = [], {}

      # grab property class
      prop = property_map[name] = _model.__dict__[name]

      # copy in default if field has explicit default value
      if prop.default != prop.sentinel:
        _pkwargs['default'] = prop.default

      # map in required and repeated kwargs
      _pkwargs['required'], _pkwargs['repeated'] = (
        prop.required, prop.repeated)

      # check for explicit field
      if _field_kwarg in prop.options:

        # grab explicit field, if any
        explicit = prop.options.get(_field_kwarg, datastructures.EMPTY)

        # explcitly setting `False` or `None` means skip this field
        if explicit is False or explicit is None:  # pragma: no cover
          continue  # continue without incrementing: skipped field

        # if it's a tuple, it's a name/args/kwargs pattern
        if not isinstance(explicit, (basestring, tuple)):
          context = (name, _model.kind(), type(explicit))
          raise TypeError('Invalid type found for explicit message field'
                          ' implementation binding - property \"%s\" of model'
                          ' \"%s\" cannot bind to field of type \"%s\".'
                          ' A basestring field name or tuple of'
                          ' (name, *args, <**kwargs>) was expected.' % context)

        elif isinstance(explicit, tuple):

          # two indicates name + args
          if len(explicit) == 2:  # name, *args
            explicit, _pargs = explicit
            _pkwargs = {}

          # three indicates name + args + kwargs
          elif len(explicit) == 3:  # name, *args, **kwargs
            explicit, _pargs, _pkwargs = explicit

        # grab explicit field (if it's not a tuple it's a basestring)
        if explicit in _builtin_fields:

          # flatten arguments, splice in ID
          if len(_pargs) > 0:
            if not isinstance(_pargs, list):
              _pargs = [i for i in _pargs]

            _field_i += 1
            _pargs.append(_field_i)
            _pargs = tuple(_pargs)
          else:
            # shortcut: replace it if there's no args
            _field_i += 1
            _pargs = (_field_i,)

          # factory field
          _model_message[name] = (
            _field_explicit_map[explicit](*_pargs, **_pkwargs))
          continue

        else:
          # raise a `ValueError` in the case of an invalid explicit field name
          raise ValueError("No such message implementation"
                           " field: \"%s\"." % name)

      # check variant by dict
      if prop.basetype == dict:
        _field_i += 1
        _model_message[name] = rpc.VariantField(_field_i)
        continue

      # check recursive submodels
      elif isinstance(prop.basetype, type(type)) and (
        issubclass(prop.basetype, model.AbstractModel)):

        # shortcut: `model.Model` for `VariantField`s
        if prop._basetype is model.Model:

          ## general, top-level `Model` means a variant field
          _field_i += 1
          _model_message[name] = rpc.VariantField(_field_i)
          continue

        # recurse - it's a model class
        _field_i += 1
        _pargs.append(prop.basetype.to_message_model())
        _pargs.append(_field_i)

        # factory
        _model_message[name] = pmessages.MessageField(*_pargs, **_pkwargs)
        continue

      # check for keys (implemented with `basestring` for now)
      elif issubclass(prop.basetype, model.AbstractKey):

        # build field and advance
        _field_i += 1
        _pargs.append(rpc.Key)
        _pargs.append(_field_i)
        _model_message[name] = pmessages.MessageField(*_pargs)
        continue

      # check for enums
      elif issubclass(prop.basetype, datastructures.BidirectionalEnum):

        # build enum class, field, and advance
        _field_i += 1
        _enum = pmessages.Enum.__metaclass__.__new__(*(
          pmessages.Enum.__metaclass__,
          prop.basetype.__name__,
          (pmessages.Enum,),
          {k: v for k, v in prop.basetype}))
        _pargs.append(_enum)
        _pargs.append(_field_i)

        if prop.default not in (
            model.Property.sentinel, None):  # pragma: no cover
          _pkwargs['default'] = (
            prop.basetype.reverse_resolve(prop.default))

        _model_message[name] = pmessages.EnumField(*_pargs, **_pkwargs)
        continue

      # check builtin basetypes
      elif prop.basetype in _builtin_basetypes:

        # build field and advance
        _field_i += 1
        _pargs.append(_field_i)
        _model_message[name] = (
          _field_basetype_map[prop.basetype](*_pargs, **_pkwargs))
        continue

      # check for builtin hook for message implementation
      elif hasattr(prop.basetype, '__message__'):

        # delegate field and advance
        _field_i += 1
        _pargs.append(_field_i)
        _model_message[name] = prop.basetype.__message__(*_pargs, **_pkwargs)
        continue

      else:  # pragma: no cover
        context = (name, _model.kind(), prop.basetype)
        raise ValueError("Could not resolve proper serialization for property"
                         " \"%s\" of model \"%s\" (found basetype \"%s\")." % (
                          context))

    # construct message class on-the-fly
    return type(_model.kind(), (pmessages.Message,), _model_message)


  ## ProtoRPCKey
  class ProtoRPCKey(KeyMixin):

    """ Adapt `Key` classes to ProtoRPC messages. """

    def to_message(self, flat=False, encoded=False):

      """ Convert a `Key` instance to a ProtoRPC `Message` instance.

          :returns: Constructed :py:class:`protorpc.Key` message object. """

      from canteen import rpc

      args = {
        'id': self.id,
        'kind': self.kind,
        'encoded': self.urlsafe()}

      if self.parent:
        if encoded:
          # indication from outer method that we should only encode
          return rpc.Key(**args)  # pragma: no cover
        args['parent'] = self.parent.to_message(not flat, flat)

      return rpc.Key(**args)

    @classmethod
    def to_message_model(cls):

      """ Return a schema for a `Key` instance in ProtoRPC `Message` form.

          :returns: Vanilla :py:class:`protorpc.Key` class. """

      from canteen import rpc
      return rpc.Key

    @classmethod
    def from_message(cls, key_message):

      """  """

      parent = cls.from_message(key_message.parent) if (
        key_message.parent
      ) else None

      # decode recursively for parent key, if specified
      return cls(key_message.kind, key_message.id, parent=parent)


  ## ProtoRPCModel
  class ProtoRPCModel(ModelMixin):

    """ Adapt Model classes to ProtoRPC messages. """

    def to_message(self, *args, **kwargs):

      """ Convert a `Model` instance to a ProtoRPC `Message` class.

          :param args: Positional arguments to pass to
            :py:meth:`Model.to_dict`.

          :param kwargs: Keyword arguments to pass to
            :py:meth:`Model.to_dict`.

          :returns: Constructed and initialized :py:class:`protorpc.Message`
            object. """

      # must import inline to avoid circular dependency
      from canteen import rpc
      from canteen import model

      values = {}
      for prop, value in self.to_dict(*args,
                                      convert_keys=False, **kwargs).items():

        # convert keys => urlsafe
        if isinstance(value, (model.Key, model.VertexKey, model.EdgeKey)):
          values[prop] = rpc.Key(
            id=value.id,
            kind=value.kind,
            encoded=value.urlsafe())
          continue

        # convert date/time/datetime => string
        if isinstance(value, (datetime.date, datetime.time, datetime.datetime)):
          values[prop] = value.isoformat()  # pragma: no cover
          continue  # pragma: no cover

        values[prop] = value  # otherwise, just set it

      if self.key:
        return self.__class__.to_message_model()(
          key=self.key.to_message(), **values)

      def _check_value(item):

        """ Checks for invalid ProtoRPC values. """

        key, value = item

        if isinstance(value, list) and len(value) == 0:
          return False  # pragma: no cover
        return True

      filtered = filter(_check_value, values.iteritems())
      return self.__class__.to_message_model()(**dict(filtered))

    @classmethod
    def to_message_model(cls):

      """ Convert a `Model` class to a ProtoRPC `Message` class. Delegates
          to :py:func:`build_message`, see docs there for exceptions raised
          (:py:exc:`TypeError` and :py:exc:`ValueError`).

          :returns: Constructed (but not initialized) dynamically-build
            :py:class:`message.Message` class corresponding to
            the current model (``cls``). """

      global _model_impl

      # check global model=>message implementation cache
      if (cls, cls.__lookup__) not in _model_impl:

        # build message class
        _model_impl[(cls, cls.__lookup__)] = build_message(cls)

      # return from cache
      return _model_impl[(cls, cls.__lookup__)]

    @classmethod
    def from_message(cls, message):

      """ DOCSTRING """

      # create an empty model, loading its key
      # (if present, which it will be if this is coming from `to_message`)
      key = cls.__keyclass__.from_message(message.key) if (
        hasattr(message, 'key') and message.key is not None) else None

      model = cls(key=key) if key else cls()

      # decode field values
      for field, value in ((
        k.name, message.get_assigned_value(k.name)) for k in (
            message.all_fields())):

        if field is not 'key':
          model[field] = value

      return model
